from pathlib import Path

from pydexpi.dexpi_classes.pydantic_classes import DexpiModel
from pydexpi.loaders.proteus_serializer.core import ParserFactoryProtocol, ProteusLoader
from pydexpi.loaders.proteus_serializer.parser_factory import ParserFactory
from pydexpi.loaders.serializer import Serializer


class ProteusSerializer(Serializer):
    """Main class for the Proteus Serializer that implements the Serializer interface."""

    def __init__(self, parser_factory: ParserFactoryProtocol = None):
        """Initialize the ProteusSerializer with a parser factory.

        Parameters
        ----------
        parser_factory : ParserFactoryProtocol, optional
            A factory to create parsers for the ProteusLoader. If not provided, a default
            ParserFactory instance will be used.
        """

        parser_factory = parser_factory or ParserFactory()
        self.proteus_loader = ProteusLoader(parser_factory)

    def save(self, model: DexpiModel, dir_path: Path, filename: str):
        """Saves a DEXPI model to a file using the ProteusLoader. Currently not implemented."""
        # The ProteusLoader does not support saving models directly.
        raise NotImplementedError("ProteusSerializer does not support saving models.")

    def load(self, dir_path: Path, filename: str) -> DexpiModel:
        """Loads a DEXPI model from a file using the ProteusLoader.

        Parameters
        ----------
        dir_path : Path
            The directory path where the file is located.
        filename : str
            The name of the file to load.

        Returns
        -------
        DexpiModel
            The loaded DEXPI model."""

        if not filename.endswith(".xml"):
            filename += ".xml"
        path = Path(dir_path) / filename

        return self.proteus_loader.load_xml_file(path)



from .pydantic_classes import (
    ActuatingElectricalFunctionShape,
    ActuatingElectricalSystemComponentShape,
    ActuatingElectricalSystemNumberLabel,
    ActuatingElectricalSystemShape,
    ActuatingFunctionShape,
    ActuatingSystemComponentShape,
    ActuatingSystemNumberLabel,
    ActuatingSystemShape,
    AttributeRepresentation,
    AttributeRepresentationType,
    Border,
    Color,
    ConnectorLine,
    CustomLabel,
    CustomSymbol,
    DashStyle,
    DeviceInformationLabel,
    Diagram,
    Ellipse,
    EllipseArc,
    EquipmentBarLabel,
    EquipmentTagNameLabel,
    FailActionLabel,
    FillStyle,
    FittingLabel,
    GraphicalElement,
    GraphicalPrimitive,
    GraphicsGroup,
    InstrumentationLoopFunctionShape,
    InstrumentationNodePosition,
    InsulationBreakLabel,
    InsulationLabel,
    InsulationSymbol,
    InsulationSymbolShape,
    Label,
    LabelShape,
    LiteralText,
    MeasuringSystemComponentShape,
    MeasuringSystemNumberLabel,
    MeasuringSystemShape,
    MPRelevanceLabel,
    NodePosition,
    NoteIdentifierLabel,
    NoteShape,
    NoteTextLabel,
    NozzleShape,
    NozzleStandardLabel,
    OffPageConnectorDescriptionLabel,
    OffPageConnectorNumberLabel,
    PipeFlowArrow,
    PipeFlowArrowShape,
    PipeOffPageConnectorShape,
    PipeSlopeSymbol,
    PipeSlopeSymbolShape,
    PipingClassBreakLabel,
    PipingComponentShape,
    PipingNetworkSegmentLabel,
    PipingNetworkSystemLabel,
    PipingNodePosition,
    Point,
    Polygon,
    PolyLine,
    ProcessInstrumentationFunctionLabel,
    ProcessInstrumentationFunctionShape,
    ProcessSignalGeneratingFunctionShape,
    PropertyBreakShape,
    QualityRelevanceLabel,
    ReducerLabel,
    ReferencedPIDNumberLabel,
    RepresentationGroup,
    RepresentationTypeGroup,
    SafetyRelevanceLabel,
    SafetyValveOrFittingLabel,
    Shape,
    ShapeCatalogue,
    ShapeUsage,
    SignalConveyingFunctionLabel,
    SignalHighHighHighLabel,
    SignalHighHighLabel,
    SignalHighLabel,
    SignalLowLabel,
    SignalLowLowLabel,
    SignalLowLowLowLabel,
    SignalOffPageConnectorShape,
    Static,
    Stroke,
    Symbol,
    SymbolShape,
    TaggedPlantItemShape,
    Text,
    TextAlignment,
    TextTemplate,
    TextTemplateFragment,
    TypicalInformationLabel,
    ValveLabel,
    VendorNameLabel,
)



"""Tests for the core module of the proteus serializer."""

import xml.etree.ElementTree as ET
from pathlib import Path
from tempfile import NamedTemporaryFile
from unittest.mock import Mock

import pytest

from pydexpi.dexpi_classes.pydantic_classes import DexpiBaseModel, DexpiModel
from pydexpi.loaders.proteus_serializer.core import (
    ErrorLevels,
    ErrorRegistry,
    ErrorRegistryEntry,
    InternalParserError,
    ModuleContext,
    ObjectRegistry,
    ParserModule,
    ProteusLoader,
)


class TestInternalParserError:
    """Test the InternalParserError exception class."""

    def test_creation(self) -> None:
        """Test creating an InternalParserError."""
        error = InternalParserError("Test error message")
        assert str(error) == "Test error message"
        assert isinstance(error, Exception)


class TestErrorLevels:
    """Test the ErrorLevels enumeration."""

    def test_all_levels_exist(self) -> None:
        """Test that all expected error levels exist."""
        assert ErrorLevels.INFO.value == "info"
        assert ErrorLevels.WARNING.value == "warning"
        assert ErrorLevels.ERROR.value == "error"
        assert ErrorLevels.CRITICAL.value == "critical"

    def test_enum_comparison(self) -> None:
        """Test that enum values can be compared."""
        assert ErrorLevels.INFO == ErrorLevels.INFO
        assert ErrorLevels.INFO != ErrorLevels.ERROR


class TestErrorRegistryEntry:
    """Test the ErrorRegistryEntry dataclass."""

    def test_minimal_entry(self) -> None:
        """Test creating an entry with minimal parameters."""
        entry = ErrorRegistryEntry(message="Test message", level=ErrorLevels.WARNING)
        assert entry.message == "Test message"
        assert entry.level == ErrorLevels.WARNING
        assert entry.proteus_id is None
        assert entry.exception is None

    def test_full_entry(self) -> None:
        """Test creating an entry with all parameters."""
        exception = ValueError("Test exception")
        entry = ErrorRegistryEntry(
            message="Test message",
            level=ErrorLevels.ERROR,
            proteus_id="test-id-123",
            exception=exception,
        )
        assert entry.message == "Test message"
        assert entry.level == ErrorLevels.ERROR
        assert entry.proteus_id == "test-id-123"
        assert entry.exception == exception


class TestErrorRegistry:
    """Test the ErrorRegistry class."""

    def test_initialization(self) -> None:
        """Test that ErrorRegistry initializes with empty error list."""
        registry = ErrorRegistry()
        assert registry.errors == []

    def test_register_error(self) -> None:
        """Test registering a normal error."""
        registry = ErrorRegistry()
        entry = ErrorRegistryEntry(
            message="Test error", level=ErrorLevels.WARNING, proteus_id="test-id"
        )

        registry.register_error(entry)

        assert len(registry.errors) == 1
        assert registry.errors[0] == entry

    def test_register_internal_parser_error(self) -> None:
        """Test that InternalParserError is reraised."""
        registry = ErrorRegistry()
        internal_error = InternalParserError("Internal error")
        entry = ErrorRegistryEntry(
            message="Internal error occurred", level=ErrorLevels.CRITICAL, exception=internal_error
        )

        with pytest.raises(InternalParserError, match="Internal error"):
            registry.register_error(entry)

        # Error should still be registered
        assert len(registry.errors) == 1

    def test_get_errors_all(self) -> None:
        """Test getting all errors without filter."""
        registry = ErrorRegistry()
        entries = [
            ErrorRegistryEntry("Info message", ErrorLevels.INFO),
            ErrorRegistryEntry("Warning message", ErrorLevels.WARNING),
            ErrorRegistryEntry("Error message", ErrorLevels.ERROR),
        ]

        for entry in entries:
            registry.register_error(entry)

        all_errors = registry.get_errors()
        assert len(all_errors) == 3
        assert all_errors == entries

    def test_get_errors_filtered(self) -> None:
        """Test getting errors filtered by level."""
        registry = ErrorRegistry()
        entries = [
            ErrorRegistryEntry("Info message", ErrorLevels.INFO),
            ErrorRegistryEntry("Warning message", ErrorLevels.WARNING),
            ErrorRegistryEntry("Error message", ErrorLevels.ERROR),
            ErrorRegistryEntry("Critical message", ErrorLevels.CRITICAL),
        ]

        for entry in entries:
            registry.register_error(entry)

        # Test filtering for ERROR and CRITICAL
        filtered_errors = registry.get_errors([ErrorLevels.ERROR, ErrorLevels.CRITICAL])
        assert len(filtered_errors) == 2
        assert all(
            error.level in [ErrorLevels.ERROR, ErrorLevels.CRITICAL] for error in filtered_errors
        )

    def test_get_errors_empty_filter(self) -> None:
        """Test getting errors with empty filter list."""
        registry = ErrorRegistry()
        entry = ErrorRegistryEntry("Test message", ErrorLevels.INFO)
        registry.register_error(entry)

        filtered_errors = registry.get_errors([])
        assert len(filtered_errors) == 0


class TestObjectRegistry:
    """Test the ObjectRegistry class."""

    def test_initialization(self) -> None:
        """Test that ObjectRegistry initializes with empty objects dict."""
        registry = ObjectRegistry()
        assert registry.objects == {}

    def test_register_and_get_object(self) -> None:
        """Test registering and retrieving an object."""
        registry = ObjectRegistry()
        test_object = {"test": "data"}
        proteus_id = "test-id-123"

        registry.register_object(proteus_id, test_object)

        retrieved = registry.get_object(proteus_id)
        assert retrieved == test_object

    def test_get_nonexistent_object(self) -> None:
        """Test getting an object that doesn't exist."""
        registry = ObjectRegistry()

        retrieved = registry.get_object("nonexistent-id")
        assert retrieved is None

    def test_register_multiple_objects(self) -> None:
        """Test registering multiple objects."""
        registry = ObjectRegistry()
        objects = {
            "id1": {"data": "object1"},
            "id2": {"data": "object2"},
            "id3": {"data": "object3"},
        }

        for proteus_id, obj in objects.items():
            registry.register_object(proteus_id, obj)

        for proteus_id, expected_obj in objects.items():
            retrieved = registry.get_object(proteus_id)
            assert retrieved == expected_obj

    def test_overwrite_object(self) -> None:
        """Test overwriting an existing object."""
        registry = ObjectRegistry()
        proteus_id = "test-id"

        registry.register_object(proteus_id, {"version": 1})
        registry.register_object(proteus_id, {"version": 2})

        retrieved = registry.get_object(proteus_id)
        assert retrieved == {"version": 2}


class TestModuleContext:
    """Test the ModuleContext class."""

    def setUp(self) -> None:
        """Set up test fixtures."""
        self.error_registry = ErrorRegistry()
        self.object_registry = ObjectRegistry()

    def test_initialization(self) -> None:
        """Test ModuleContext initialization."""
        self.setUp()
        context = ModuleContext(
            element_stack=["root", "child"],
            id_stack=["root-id", "child-id"],
            error_registry=self.error_registry,
            object_registry=self.object_registry,
        )

        assert context.element_stack == ["root", "child"]
        assert context.id_stack == ["root-id", "child-id"]
        assert context.error_registry == self.error_registry
        assert context.object_registry == self.object_registry

    def test_get_updated_context(self) -> None:
        """Test creating updated context with new element."""
        self.setUp()
        context = ModuleContext(
            element_stack=["root"],
            id_stack=["root-id"],
            error_registry=self.error_registry,
            object_registry=self.object_registry,
        )

        # Create XML element with ID
        element = ET.Element("child")
        element.set("ID", "child-id")

        updated_context = context.get_updated_context(element)

        assert updated_context.element_stack == ["root", "child"]
        assert updated_context.id_stack == ["root-id", "child-id"]
        assert updated_context.error_registry == self.error_registry
        assert updated_context.object_registry == self.object_registry

    def test_get_updated_context_no_id(self) -> None:
        """Test creating updated context with element that has no ID."""
        self.setUp()
        context = ModuleContext(
            element_stack=["root"],
            id_stack=["root-id"],
            error_registry=self.error_registry,
            object_registry=self.object_registry,
        )

        element = ET.Element("child")  # No ID attribute

        updated_context = context.get_updated_context(element)

        assert updated_context.element_stack == ["root", "child"]
        assert updated_context.id_stack == ["root-id", None]

    def test_get_last_id_success(self) -> None:
        """Test getting the last non-None ID from stack."""
        self.setUp()
        context = ModuleContext(
            element_stack=["root", "child", "grandchild"],
            id_stack=["root-id", "child-id", None],
            error_registry=self.error_registry,
            object_registry=self.object_registry,
        )

        last_id = context.get_last_id()
        assert last_id == "child-id"

    def test_get_last_id_all_none(self) -> None:
        """Test getting last ID when all IDs are None."""
        self.setUp()
        context = ModuleContext(
            element_stack=["root", "child"],
            id_stack=[None, None],
            error_registry=self.error_registry,
            object_registry=self.object_registry,
        )

        with pytest.raises(ValueError, match="ID stack is empty or contains only None values"):
            context.get_last_id()

    def test_get_last_id_empty_stack(self) -> None:
        """Test getting last ID when stack is empty."""
        self.setUp()
        context = ModuleContext(
            element_stack=[],
            id_stack=[],
            error_registry=self.error_registry,
            object_registry=self.object_registry,
        )

        with pytest.raises(ValueError, match="ID stack is empty or contains only None values"):
            context.get_last_id()

    def test_register_error_with_explicit_id(self) -> None:
        """Test registering error with explicit proteus_id."""
        self.setUp()
        context = ModuleContext(
            element_stack=["root"],
            id_stack=["root-id"],
            error_registry=self.error_registry,
            object_registry=self.object_registry,
        )

        context.register_error("Test error", ErrorLevels.WARNING, proteus_id="explicit-id")

        errors = self.error_registry.get_errors()
        assert len(errors) == 1
        assert errors[0].proteus_id == "explicit-id"

    def test_register_error_with_inferred_id(self) -> None:
        """Test registering error with ID inferred from context."""
        self.setUp()
        context = ModuleContext(
            element_stack=["root", "child"],
            id_stack=["root-id", "child-id"],
            error_registry=self.error_registry,
            object_registry=self.object_registry,
        )

        context.register_error("Test error", ErrorLevels.WARNING)

        errors = self.error_registry.get_errors()
        assert len(errors) == 1
        assert errors[0].proteus_id == "child-id"

    def test_register_error_no_id_available(self) -> None:
        """Test registering error when no ID is available."""
        self.setUp()
        context = ModuleContext(
            element_stack=["root"],
            id_stack=[None],
            error_registry=self.error_registry,
            object_registry=self.object_registry,
        )

        context.register_error("Test error", ErrorLevels.WARNING)

        errors = self.error_registry.get_errors()
        assert len(errors) == 1
        assert errors[0].proteus_id is None


class ConcreteParserModule(ParserModule):
    """Concrete implementation of ParserModule for testing."""

    def __init__(self, context: ModuleContext, return_value: str = "test_result") -> None:
        """Initialize concrete parser module.

        Parameters
        ----------
        context : ModuleContext
            The module context
        return_value : str, optional
            Value to return from compositional_pass, by default "test_result"
        """
        super().__init__(context)
        self.return_value = return_value

    def compositional_pass(self) -> str:
        """Return test result from compositional pass.

        Returns
        -------
        str
            Test result value
        """
        return self.return_value


class TestParserModule:
    """Test the ParserModule abstract base class."""

    def setUp(self) -> None:
        """Set up test fixtures."""
        self.error_registry = ErrorRegistry()
        self.object_registry = ObjectRegistry()
        self.context = ModuleContext(
            element_stack=["root"],
            id_stack=["root-id"],
            error_registry=self.error_registry,
            object_registry=self.object_registry,
        )

    def test_initialization(self) -> None:
        """Test ParserModule initialization."""
        self.setUp()
        module = ConcreteParserModule(self.context)

        assert module.context == self.context
        assert module.submodules == []

    def test_compositional_pass(self) -> None:
        """Test compositional pass method."""
        self.setUp()
        module = ConcreteParserModule(self.context, "custom_result")

        result = module.compositional_pass()
        assert result == "custom_result"

    def test_reference_pass_no_submodules(self) -> None:
        """Test reference pass with no submodules."""
        self.setUp()
        module = ConcreteParserModule(self.context)

        # Should not raise any exceptions
        module.reference_pass()

    def test_reference_pass_with_submodules(self) -> None:
        """Test reference pass with submodules."""
        self.setUp()
        module = ConcreteParserModule(self.context)

        # Create mock submodules
        submodule1 = Mock(spec=ParserModule)
        submodule2 = Mock(spec=ParserModule)

        module.register_submodule(submodule1)
        module.register_submodule(submodule2)

        module.reference_pass()

        submodule1.reference_pass.assert_called_once()
        submodule2.reference_pass.assert_called_once()

    def test_control_pass_with_submodules(self) -> None:
        """Test control pass with submodules."""
        self.setUp()
        module = ConcreteParserModule(self.context)

        # Create mock submodules
        submodule1 = Mock(spec=ParserModule)
        submodule2 = Mock(spec=ParserModule)

        module.register_submodule(submodule1)
        module.register_submodule(submodule2)

        module.control_pass()

        submodule1.control_pass.assert_called_once()
        submodule2.control_pass.assert_called_once()

    def test_register_submodule(self) -> None:
        """Test registering a single submodule."""
        self.setUp()
        module = ConcreteParserModule(self.context)
        submodule = ConcreteParserModule(self.context)

        module.register_submodule(submodule)

        assert len(module.submodules) == 1
        assert module.submodules[0] == submodule

    def test_register_submodule_list(self) -> None:
        """Test registering multiple submodules at once."""
        self.setUp()
        module = ConcreteParserModule(self.context)
        submodules = [
            ConcreteParserModule(self.context),
            ConcreteParserModule(self.context),
            ConcreteParserModule(self.context),
        ]

        module.register_submodule_list(submodules)

        assert len(module.submodules) == 3
        assert module.submodules == submodules

    def test_register_error(self) -> None:
        """Test registering an error through the module."""
        self.setUp()
        module = ConcreteParserModule(self.context)

        module.register_error("Module error", ErrorLevels.ERROR)

        errors = self.error_registry.get_errors()
        assert len(errors) == 1
        assert errors[0].message == "Module error"
        assert errors[0].level == ErrorLevels.ERROR

    def test_register_object(self) -> None:
        """Test registering an object through the module."""
        self.setUp()
        module = ConcreteParserModule(self.context)
        test_object = Mock(spec=DexpiBaseModel)

        module.register_object("test-id", test_object)

        retrieved = self.object_registry.get_object("test-id")
        assert retrieved == test_object

    def test_get_object_from_registry(self) -> None:
        """Test getting an object from registry through the module."""
        self.setUp()
        module = ConcreteParserModule(self.context)
        test_object = Mock(spec=DexpiBaseModel)

        self.object_registry.register_object("test-id", test_object)

        retrieved = module.get_object_from_registry("test-id")
        assert retrieved == test_object

    def test_get_nonexistent_object_from_registry(self) -> None:
        """Test getting nonexistent object returns None."""
        self.setUp()
        module = ConcreteParserModule(self.context)

        retrieved = module.get_object_from_registry("nonexistent-id")
        assert retrieved is None


class TestProteusLoader:
    """Test the ProteusLoader class."""

    def test_initialization(self) -> None:
        """Test ProteusLoader initialization."""
        factory = Mock()
        loader = ProteusLoader(factory)

        assert loader.parser_factory == factory
        assert loader.error_registry is None
        assert loader.object_registry is None

    def test_reset_registries(self) -> None:
        """Test resetting registries creates new instances."""
        factory = Mock()
        loader = ProteusLoader(factory)

        loader.reset_registries()

        assert isinstance(loader.error_registry, ErrorRegistry)
        assert isinstance(loader.object_registry, ObjectRegistry)
        assert loader.error_registry.errors == []
        assert loader.object_registry.objects == {}

    def test_load_xmlstring_success(self) -> None:
        """Test successful XML string loading."""
        # Create mock factory and parser
        factory = Mock()
        parser = Mock()
        expected_model = Mock(spec=DexpiModel)

        factory.make_plant_model_parser.return_value = parser
        parser.compositional_pass.return_value = expected_model

        loader = ProteusLoader(factory)

        xml_string = "<root><PlantModel></PlantModel></root>"
        result = loader.load_xmlstring(xml_string)

        # Verify the loading sequence
        factory.make_plant_model_parser.assert_called_once()
        parser.compositional_pass.assert_called_once()
        parser.reference_pass.assert_called_once()
        parser.control_pass.assert_called_once()

        assert result == expected_model
        assert isinstance(loader.error_registry, ErrorRegistry)
        assert isinstance(loader.object_registry, ObjectRegistry)

    def test_load_xmlstring_invalid_xml(self) -> None:
        """Test loading invalid XML string raises exception."""
        factory = Mock()
        loader = ProteusLoader(factory)

        invalid_xml = "<root><unclosed>"

        with pytest.raises(ET.ParseError):
            loader.load_xmlstring(invalid_xml)

    def test_load_xml_file_success(self) -> None:
        """Test successful XML file loading."""
        # Create mock factory and parser
        factory = Mock()
        parser = Mock()
        expected_model = Mock(spec=DexpiModel)

        factory.make_plant_model_parser.return_value = parser
        parser.compositional_pass.return_value = expected_model

        loader = ProteusLoader(factory)

        # Create temporary XML file
        xml_content = "<root><PlantModel></PlantModel></root>"

        with NamedTemporaryFile(mode="w", suffix=".xml", delete=False) as tmp_file:
            tmp_file.write(xml_content)
            tmp_file_path = Path(tmp_file.name)

        try:
            result = loader.load_xml_file(tmp_file_path)

            # Verify the loading sequence
            factory.make_plant_model_parser.assert_called_once()
            parser.compositional_pass.assert_called_once()
            parser.reference_pass.assert_called_once()
            parser.control_pass.assert_called_once()

            assert result == expected_model
        finally:
            tmp_file_path.unlink()  # Clean up temp file

    def test_load_xml_file_nonexistent(self) -> None:
        """Test loading nonexistent file raises exception."""
        factory = Mock()
        loader = ProteusLoader(factory)

        nonexistent_path = Path("nonexistent_file.xml")

        with pytest.raises(FileNotFoundError):
            loader.load_xml_file(nonexistent_path)

    def test_get_error_registry(self) -> None:
        """Test getting error registry."""
        factory = Mock()
        loader = ProteusLoader(factory)

        loader.reset_registries()
        error_registry = loader.get_error_registry()

        assert error_registry == loader.error_registry
        assert isinstance(error_registry, ErrorRegistry)

    def test_get_error_registry_before_reset(self) -> None:
        """Test getting error registry before reset returns None."""
        factory = Mock()
        loader = ProteusLoader(factory)

        error_registry = loader.get_error_registry()
        assert error_registry is None

    def test_context_creation_in_load_xmlstring(self) -> None:
        """Test that proper context is created during XML string loading."""
        factory = Mock()
        parser = Mock()
        expected_model = Mock(spec=DexpiModel)

        factory.make_plant_model_parser.return_value = parser
        parser.compositional_pass.return_value = expected_model

        loader = ProteusLoader(factory)

        xml_string = "<TestRoot><PlantModel></PlantModel></TestRoot>"
        loader.load_xmlstring(xml_string)

        # Check that factory was called with correct context
        call_args = factory.make_plant_model_parser.call_args
        context = call_args[0][0]  # First argument
        element = call_args[0][1]  # Second argument

        assert isinstance(context, ModuleContext)
        assert context.element_stack == ["TestRoot"]
        assert context.id_stack == [None]
        assert isinstance(context.error_registry, ErrorRegistry)
        assert isinstance(context.object_registry, ObjectRegistry)
        assert element.tag == "TestRoot"


# Integration test to verify all components work together
class TestCoreIntegration:
    """Integration tests for core module components."""

    def test_full_workflow_simulation(self) -> None:
        """Test a complete workflow simulation with all core components."""
        # Create registries
        error_registry = ErrorRegistry()
        object_registry = ObjectRegistry()

        # Create context
        context = ModuleContext(
            element_stack=["root"],
            id_stack=["root-id"],
            error_registry=error_registry,
            object_registry=object_registry,
        )

        # Create parser module
        module = ConcreteParserModule(context)

        # Register some test objects
        test_obj1 = Mock(spec=DexpiBaseModel)
        test_obj2 = Mock(spec=DexpiBaseModel)

        module.register_object("obj1", test_obj1)
        module.register_object("obj2", test_obj2)

        # Register some errors
        module.register_error("Warning message", ErrorLevels.WARNING)
        module.register_error("Error message", ErrorLevels.ERROR, proteus_id="custom-id")

        # Verify objects are registered
        assert module.get_object_from_registry("obj1") == test_obj1
        assert module.get_object_from_registry("obj2") == test_obj2

        # Verify errors are registered
        errors = error_registry.get_errors()
        assert len(errors) == 2
        assert errors[0].level == ErrorLevels.WARNING
        assert errors[0].proteus_id == "root-id"  # Inferred from context
        assert errors[1].level == ErrorLevels.ERROR
        assert errors[1].proteus_id == "custom-id"  # Explicit ID

        # Test compositional pass
        result = module.compositional_pass()
        assert result == "test_result"

    def test_context_propagation(self) -> None:
        """Test that context is properly propagated through updates."""
        error_registry = ErrorRegistry()
        object_registry = ObjectRegistry()

        # Create initial context
        context = ModuleContext(
            element_stack=["root"],
            id_stack=["root-id"],
            error_registry=error_registry,
            object_registry=object_registry,
        )

        # Create XML elements to simulate hierarchy
        child_element = ET.Element("child")
        child_element.set("ID", "child-id")

        grandchild_element = ET.Element("grandchild")
        # No ID for grandchild

        # Update context through hierarchy
        child_context = context.get_updated_context(child_element)
        grandchild_context = child_context.get_updated_context(grandchild_element)

        # Verify context stack progression
        assert context.element_stack == ["root"]
        assert context.id_stack == ["root-id"]

        assert child_context.element_stack == ["root", "child"]
        assert child_context.id_stack == ["root-id", "child-id"]

        assert grandchild_context.element_stack == ["root", "child", "grandchild"]
        assert grandchild_context.id_stack == ["root-id", "child-id", None]

        # Verify shared registries
        assert child_context.error_registry == error_registry
        assert child_context.object_registry == object_registry
        assert grandchild_context.error_registry == error_registry
        assert grandchild_context.object_registry == object_registry

        # Test ID resolution at different levels
        assert context.get_last_id() == "root-id"
        assert child_context.get_last_id() == "child-id"
        assert grandchild_context.get_last_id() == "child-id"  # Should skip None



from .pydantic_classes import (
    AngleBallValve,
    AngleGlobeValve,
    AnglePlugValve,
    AngleValve,
    BallValve,
    BlindFlange,
    BreatherValve,
    ButterflyValve,
    CheckValve,
    ClampedFlangeCoupling,
    Compensator,
    ConicalStrainer,
    CustomCheckValve,
    CustomInlinePrimaryElement,
    CustomOperatedValve,
    CustomPipeFitting,
    CustomPipingComponent,
    CustomSafetyValveOrFitting,
    DirectPipingConnection,
    ElectromagneticFlowMeter,
    FlameArrestor,
    Flange,
    FlangedConnection,
    FlowInPipeOffPageConnector,
    FlowMeasuringElement,
    FlowNozzle,
    FlowOutPipeOffPageConnector,
    Funnel,
    GateValve,
    GlobeCheckValve,
    GlobeValve,
    Hose,
    IlluminatedSightGlass,
    InLineMixer,
    InlinePrimaryElement,
    LineBlind,
    MassFlowMeasuringElement,
    NeedleValve,
    OperatedValve,
    Penetration,
    Pipe,
    PipeCoupling,
    PipeFitting,
    PipeFlangeSpacer,
    PipeFlangeSpade,
    PipeOffPageConnector,
    PipeOffPageConnectorObjectReference,
    PipeOffPageConnectorReference,
    PipeOffPageConnectorReferenceByNumber,
    PipeReducer,
    PipeTee,
    PipingComponent,
    PipingConnection,
    PipingNetworkSegment,
    PipingNetworkSegmentItem,
    PipingNetworkSystem,
    PipingNode,
    PipingNodeOwner,
    PipingSourceItem,
    PipingTargetItem,
    PlugValve,
    PositiveDisplacementFlowMeter,
    PropertyBreak,
    RestrictionOrifice,
    RuptureDisc,
    SafetyValveOrFitting,
    SightGlass,
    Silencer,
    SpringLoadedAngleGlobeSafetyValve,
    SpringLoadedGlobeSafetyValve,
    SteamTrap,
    StraightwayValve,
    Strainer,
    SwingCheckValve,
    TurbineFlowMeter,
    VariableAreaFlowMeter,
    VentilationDevice,
    VenturiTube,
    VolumeFlowMeasuringElement,
)



import copy

import pytest

from pydexpi.dexpi_classes import piping
from pydexpi.toolkits import piping_toolkit as pt


def test_segment_ends_in_connection(simple_pns_factory):
    """Test evaluating if a segment ends in a connection."""
    segment = simple_pns_factory()
    assert pt.segment_ends_in_connection(segment) is False
    assert pt.segment_ends_in_connection(segment, as_source=True) is True


def test_segments_are_connected(simple_pns_factory):
    """Test evaluating if two segments are connected"""
    segment = simple_pns_factory()
    second_segment = simple_pns_factory()  # use factory to get a new instance
    assert pt.segments_are_connected(segment, second_segment) is False
    pt.connect_piping_network_segment(
        second_segment, segment.items[-1], connector_node_index=1, as_source=True
    )
    assert (
        pt.piping_network_segment_validity_check(second_segment)[0] == pt.PipingValidityCode.VALID
    )
    assert pt.segments_are_connected(segment, second_segment) is True


def test_connect_piping_network_segment(simple_pns_factory):
    """Test the reconnecting of a piping network segment. Depends on validity
    check"""
    segment = simple_pns_factory()
    new_destination_1 = piping.BallValve(nodes=[piping.PipingNode() for i in range(2)])
    new_destination_2 = copy.deepcopy(new_destination_1)
    # Cant reconnect segment that ends in an Item
    with pytest.raises(ValueError):
        pt.connect_piping_network_segment(segment, new_destination_1, connector_node_index=0)
    # Try normal reconnection
    pt.connect_piping_network_segment(
        segment, new_destination_1, connector_node_index=1, as_source=True
    )
    assert pt.piping_network_segment_validity_check(segment)[0] == pt.PipingValidityCode.VALID
    # Reconnection not possible unless force reconnect is set to True
    with pytest.raises(pt.DexpiConnectionException):
        pt.connect_piping_network_segment(
            segment,
            new_destination_2,
            connector_node_index=1,
            as_source=True,
        )
    pt.connect_piping_network_segment(
        segment,
        new_destination_2,
        connector_node_index=1,
        as_source=True,
        force_reconnect=True,
    )
    assert pt.piping_network_segment_validity_check(segment)[0] == pt.PipingValidityCode.VALID
    # Reconnection to 1, but without a node this time
    pt.connect_piping_network_segment(
        segment,
        new_destination_2,
        as_source=True,
        force_reconnect=True,
    )
    assert pt.piping_network_segment_validity_check(segment)[0] == pt.PipingValidityCode.VALID


def test_construct_new_segment():
    """Construct simple piping network segment with three pipes and valves and
    see if some of the connections are implemented correctly.

    """
    valves = [piping.BallValve(nodes=[piping.PipingNode() for i in range(2)]) for j in range(3)]
    pipes = [piping.Pipe() for i in range(3)]
    segment = pt.construct_new_segment(
        valves,
        pipes,
        target_connector_item=valves[-1],
        target_connector_node_index=1,
    )
    valves_use_items = [
        piping.BallValve(nodes=[piping.PipingNode() for i in range(2)]) for j in range(3)
    ]
    pipes_use_items = [piping.Pipe() for i in range(3)]
    segment_use_items = pt.construct_new_segment(
        valves_use_items,
        pipes_use_items,
        target_connector_item=valves_use_items[-1],
        connectivity_convention=pt.PipingConnectionConvention.USE_ITEMS,
    )
    # See if the second pipe has the right destination
    assert segment.connections[1].targetItem == segment.items[1]
    assert segment.connections[1].targetNode == segment.items[1].nodes[0]
    # See if the second pipe has the right source
    assert segment_use_items.connections[1].sourceItem == segment_use_items.items[0]
    assert segment_use_items.connections[1].sourceNode is None
    # See if the final pipe has the right destination
    assert segment.connections[-1].targetItem == segment.items[-1]
    assert segment.connections[-1].targetNode == segment.items[-1].nodes[0]
    assert segment_use_items.connections[-1].targetItem == segment_use_items.items[-1]
    assert segment_use_items.connections[-1].targetNode is None

    # See if the segment references the right object as destination
    assert segment.targetItem == valves[-1]
    assert segment.targetNode == valves[-1].nodes[1]
    assert segment_use_items.targetItem == valves_use_items[-1]
    assert segment_use_items.targetNode is None


def test_construct_new_segment_already_connected():
    """Test constructing a new segment with items that are already connected."""
    # Create test components
    valves = [piping.BallValve(nodes=[piping.PipingNode() for i in range(2)]) for j in range(3)]
    pipes = [piping.Pipe() for i in range(3)]

    # Pre-connect components using nodes
    pipes[0].sourceItem = valves[0]
    pipes[0].sourceNode = valves[0].nodes[1]
    pipes[0].targetItem = valves[1]
    pipes[0].targetNode = valves[1].nodes[0]

    pipes[1].sourceItem = valves[1]
    pipes[1].sourceNode = valves[1].nodes[1]
    pipes[1].targetItem = valves[2]
    pipes[1].targetNode = valves[2].nodes[0]

    # Test with USE_NODES convention
    segment = pt.construct_new_segment_already_connected(
        valves,
        pipes[:2],  # Only use first two pipes
        target_connector_item=valves[-1],
        target_connector_node_index=1,
    )

    # Verify connections are preserved
    assert segment.connections[0].sourceItem == valves[0]
    assert segment.connections[0].sourceNode == valves[0].nodes[1]
    assert segment.connections[0].targetItem == valves[1]
    assert segment.connections[0].targetNode == valves[1].nodes[0]

    assert segment.connections[1].sourceItem == valves[1]
    assert segment.connections[1].sourceNode == valves[1].nodes[1]
    assert segment.connections[1].targetItem == valves[2]
    assert segment.connections[1].targetNode == valves[2].nodes[0]

    # Test with USE_ITEMS convention
    valves_use_items = [piping.BallValve() for j in range(3)]
    pipes_use_items = [piping.Pipe() for i in range(3)]

    # Pre-connect components without nodes
    pipes_use_items[0].sourceItem = valves_use_items[0]
    pipes_use_items[0].targetItem = valves_use_items[1]

    pipes_use_items[1].sourceItem = valves_use_items[1]
    pipes_use_items[1].targetItem = valves_use_items[2]

    segment_use_items = pt.construct_new_segment_already_connected(
        valves_use_items,
        pipes_use_items[:2],
        target_connector_item=valves_use_items[-1],
        connectivity_convention=pt.PipingConnectionConvention.USE_ITEMS,
    )

    # Verify USE_ITEMS connections
    assert segment_use_items.connections[0].sourceItem == valves_use_items[0]
    assert segment_use_items.connections[0].sourceNode is None
    assert segment_use_items.connections[0].targetItem == valves_use_items[1]
    assert segment_use_items.connections[0].targetNode is None

    # Test error cases
    # Case 1: Disconnected items
    bad_pipes = copy.deepcopy(pipes[:2])
    bad_pipes[1].targetItem = None
    with pytest.raises(ValueError):
        pt.construct_new_segment_already_connected(
            valves,
            bad_pipes,
            target_connector_item=valves[-1],
        )

    # Case 2: Wrong connection order
    bad_pipes = copy.deepcopy(pipes[:2])
    bad_pipes[0].sourceItem = valves[1]
    bad_pipes[0].targetItem = valves[2]
    with pytest.raises(ValueError):
        pt.construct_new_segment_already_connected(
            valves,
            bad_pipes,
            target_connector_item=valves[-1],
        )


def test_insert_item_to_segment(simple_pns_factory):
    """Test inserting an item to a simple piping network segment"""
    segment = simple_pns_factory()
    # Insert an item into an empty segment
    new_item = piping.BallValve(nodes=[piping.PipingNode() for i in range(2)])
    new_connection = piping.Pipe()
    the_segment = piping.PipingNetworkSegment()
    pt.insert_item_to_segment(
        the_segment,
        0,
        new_item,
        new_connection,
        item_target_node_index=0,
        item_source_node_index=1,
        insert_before=True,
    )
    assert pt.piping_network_segment_validity_check(the_segment)[0] == pt.PipingValidityCode.VALID
    new_item = piping.FlowInPipeOffPageConnector()
    new_connection = piping.Pipe()
    the_segment = piping.PipingNetworkSegment()
    pt.insert_item_to_segment(the_segment, 0, new_item, new_connection, insert_before=True)
    assert pt.piping_network_segment_validity_check(the_segment)[0] == pt.PipingValidityCode.VALID
    new_item = piping.FlowInPipeOffPageConnector()
    # Insert an item into a segment with just one item
    new_item = piping.BallValve(nodes=[piping.PipingNode() for i in range(2)])
    new_connection = piping.Pipe()
    the_segment = piping.PipingNetworkSegment()
    pt.append_item_to_unconnected_segment(the_segment, piping.BallValve())
    pt.insert_item_to_segment(
        the_segment,
        0,
        new_item,
        new_connection,
        item_target_node_index=0,
        item_source_node_index=1,
        insert_before=True,
    )
    assert pt.piping_network_segment_validity_check(the_segment)[0] == pt.PipingValidityCode.VALID
    # Insert an item into a segment with just one connection
    new_item = piping.BallValve(nodes=[piping.PipingNode() for i in range(2)])
    new_connection = piping.Pipe()
    the_segment = piping.PipingNetworkSegment()
    pt.append_connection_to_unconnected_segment(the_segment, piping.Pipe())
    pt.insert_item_to_segment(
        the_segment,
        0,
        new_item,
        new_connection,
        item_target_node_index=0,
        item_source_node_index=1,
    )
    assert pt.piping_network_segment_validity_check(the_segment)[0] == pt.PipingValidityCode.VALID
    # Try inserting after a connection not yet in the segment
    with pytest.raises(ValueError):
        pt.insert_item_to_segment(
            segment,
            copy.deepcopy(new_item),
            new_item,
            new_connection,
            item_target_node_index=0,
            item_source_node_index=1,
        )
    # Try inserting in some different positions
    pt.insert_item_to_segment(
        segment,
        1,
        new_item,
        new_connection,
        item_target_node_index=0,
        item_source_node_index=1,
        insert_before=True,
    )
    assert pt.piping_network_segment_validity_check(segment)[0] == pt.PipingValidityCode.VALID
    pt.insert_item_to_segment(segment, -2, copy.deepcopy(new_item), copy.deepcopy(new_connection))
    assert pt.piping_network_segment_validity_check(segment)[0] == pt.PipingValidityCode.VALID
    pt.insert_item_to_segment(
        segment,
        new_item,
        copy.deepcopy(new_item),
        copy.deepcopy(new_connection),
        item_target_node_index=0,
        item_source_node_index=1,
    )
    assert pt.piping_network_segment_validity_check(segment)[0] == pt.PipingValidityCode.VALID
    # Try adding objects that are already part of the segment
    with pytest.raises(ValueError):
        pt.insert_item_to_segment(
            segment,
            1,
            new_item,
            new_connection,
            item_target_node_index=0,
            item_source_node_index=1,
        )
    # Try adding items to some invalid segments
    invalid_segment = simple_pns_factory()
    del invalid_segment.connections[1]
    new_item = piping.BallValve()
    new_connection = piping.Pipe()
    with pytest.raises(pt.DexpiCorruptPipingSegmentException):
        pt.insert_item_to_segment(invalid_segment, 1, new_item, new_connection, insert_before=True)


def test_append_item_to_unconnected_segment(simple_pns_factory):
    """Test appending an item to a simple piping network segment"""
    segment = simple_pns_factory()
    # Append an item to an empty segment
    new_item = piping.BallValve(nodes=[piping.PipingNode() for i in range(2)])
    new_segment = piping.PipingNetworkSegment()
    pt.append_item_to_unconnected_segment(new_segment, new_item, 1)
    assert pt.piping_network_segment_validity_check(segment)[0] == pt.PipingValidityCode.VALID
    # Try appending an internal item, which is not allowed
    with pytest.raises(ValueError):
        pt.append_item_to_unconnected_segment(segment, segment.items[0], 1, insert_before=True)
    # Try appending a pipe at an end that already has a pipe
    new_item = piping.BallValve(nodes=[piping.PipingNode() for i in range(2)])
    with pytest.raises(ValueError):
        pt.append_item_to_unconnected_segment(segment, new_item, 0)
    # Append an item normally and see if the resulting segment is valid
    pt.append_item_to_unconnected_segment(segment, new_item, 1, 0, insert_before=True)
    assert pt.piping_network_segment_validity_check(segment)[0] == pt.PipingValidityCode.VALID


def test_append_connection_to_unconnected_segment(simple_pns_factory):
    """Test appending a connection to a simple piping network segment"""
    segment = simple_pns_factory()
    # Append a connection to an empty segment
    new_connection = piping.Pipe()
    new_segment = piping.PipingNetworkSegment()
    pt.append_connection_to_unconnected_segment(new_segment, new_connection, 1)
    assert pt.piping_network_segment_validity_check(segment)[0] == pt.PipingValidityCode.VALID
    # Try appending an internal connection, which is not allowed
    with pytest.raises(ValueError):
        pt.append_connection_to_unconnected_segment(segment, segment.connections[0], 1)
    # Try appending a pipe at an end that already has a pipe
    with pytest.raises(ValueError):
        pt.append_connection_to_unconnected_segment(segment, piping.Pipe(), 0, insert_before=True)
    # Append a pipe normally and see if the resulting segment is valid
    pt.append_connection_to_unconnected_segment(segment, piping.Pipe(), 1)
    assert pt.piping_network_segment_validity_check(segment)[0] == pt.PipingValidityCode.VALID


def test_find_final_connection(simple_pns_factory):
    """Test if the final (first) connection can be found on the simple pns"""
    segment = simple_pns_factory()
    final_connection = pt.find_final_connection(segment)
    first_connection = pt.find_final_connection(segment, as_source=True)
    assert final_connection == segment.connections[-1]
    assert first_connection == segment.connections[0]

    # Damage pns to see if this is handled
    segment.targetNode = None
    segment.targetItem = None
    with pytest.raises(pt.DexpiCorruptPipingSegmentException):
        pt.find_final_connection(segment)


def test_get_unconnected_piping_segments(simple_conceptual_model_factory):
    """Construct a simple Conceptual model and see if unconnected connections
    are identified"""
    model = simple_conceptual_model_factory()
    outlets = pt.get_unconnected_piping_segments(model)
    inlets = pt.get_unconnected_piping_segments(model, as_source=True)
    assert len(outlets) == 1
    assert len(inlets) == 1
    assert outlets[0] == model.pipingNetworkSystems[1].segments[0]
    assert inlets[0] == model.pipingNetworkSystems[0].segments[0]


def test_traverse_items_and_connections():
    """Test traversing through connected items and connections."""
    # Create test components
    items = [piping.BallValve(nodes=[piping.PipingNode() for i in range(2)]) for _ in range(2)]
    connections = [piping.Pipe() for _ in range(5)]
    items.append(piping.PipeTee(nodes=[piping.PipingNode() for i in range(3)]))

    # Connect components
    connections[0].targetItem = items[0]
    connections[0].targetNode = items[0].nodes[0]

    connections[1].sourceItem = items[0]
    connections[1].sourceNode = items[0].nodes[1]
    connections[1].targetItem = items[1]
    connections[1].targetNode = items[1].nodes[0]

    connections[2].sourceItem = items[1]
    connections[2].sourceNode = items[1].nodes[1]
    connections[2].targetItem = items[2]
    connections[2].targetNode = items[2].nodes[0]

    connections[3].sourceItem = items[2]
    connections[3].sourceNode = items[2].nodes[1]

    connections[4].sourceItem = items[2]
    connections[4].sourceNode = items[2].nodes[2]

    traversal = pt.traverse_items_and_connections(
        items, connections, connections[0], lambda x: x == connections[4]
    )
    # Check if the traversal is correct
    assert traversal == [
        connections[0],
        items[0],
        connections[1],
        items[1],
        connections[2],
        items[2],
        connections[4],
    ]

    # Check if the traversal is correct in reverse order
    traversal = pt.traverse_items_and_connections(
        items, connections, connections[4], lambda x: x == connections[0], reverse=True
    )
    assert traversal == [
        connections[4],
        items[2],
        connections[2],
        items[1],
        connections[1],
        items[0],
        connections[0],
    ]

    # Check traversing from an item
    traversal = pt.traverse_items_and_connections(
        items, connections, items[0], lambda x: x == connections[4]
    )
    assert traversal == [
        items[0],
        connections[1],
        items[1],
        connections[2],
        items[2],
        connections[4],
    ]

    # Check traversing from an item in reverse order
    traversal = pt.traverse_items_and_connections(
        items, connections, items[2], lambda x: x == connections[0], reverse=True
    )
    assert traversal == [
        items[2],
        connections[2],
        items[1],
        connections[1],
        items[0],
        connections[0],
    ]

    # Check specifying a connection directly as end condition
    traversal = pt.traverse_items_and_connections(items, connections, items[0], connections[4])
    assert traversal == [
        items[0],
        connections[1],
        items[1],
        connections[2],
        items[2],
        connections[4],
    ]

    # Some more edge cases
    # Case 1: Traversing a single item
    single_item = piping.BallValve(nodes=[piping.PipingNode() for i in range(2)])
    traversal = pt.traverse_items_and_connections(
        [single_item], [], single_item, lambda x: x == single_item
    )
    assert traversal == [single_item]

    # Case 2: Traversing a single connection
    single_connection = piping.Pipe()
    traversal = pt.traverse_items_and_connections(
        [], [single_connection], single_connection, lambda x: x == single_connection
    )
    assert traversal == [single_connection]

    # Case 3: A little more complex end condition
    def end_condition(x):
        if isinstance(x, piping.PipingConnection):
            if x.targetItem not in items:
                return True
        return False

    traversal = pt.traverse_items_and_connections(items, connections, connections[0], end_condition)
    assert traversal == [
        connections[0],
        items[0],
        connections[1],
        items[1],
        connections[2],
        items[2],
        connections[3],
    ]

    # Test error cases
    # Case 1: Empty items list
    with pytest.raises(pt.PipingTraversalException):
        list(
            pt.traverse_items_and_connections(
                [], connections, connections[0], lambda x: x == connections[4]
            )
        )

    # Case 2: Empty connections list
    with pytest.raises(pt.PipingTraversalException):
        list(
            pt.traverse_items_and_connections(
                items, [], connections[0], lambda x: x == connections[4]
            )
        )

    # Case 3: Starting element not in lists
    extra_connection = piping.Pipe()
    with pytest.raises(pt.PipingTraversalException):
        list(
            pt.traverse_items_and_connections(
                items, connections, extra_connection, lambda x: x == connections[4]
            )
        )

    # Case 4: Target condition never met
    with pytest.raises(pt.PipingTraversalException):
        list(pt.traverse_items_and_connections(items, connections, connections[0], lambda x: False))

    # Case 5: Disconnected items
    bad_connections = copy.deepcopy(connections)
    bad_connections[2].targetItem = None
    with pytest.raises(pt.PipingTraversalException):
        list(
            pt.traverse_items_and_connections(
                items, bad_connections, connections[0], lambda x: x == connections[4]
            )
        )

    # Case 6: Loop in connections
    loop_connections = copy.deepcopy(connections)
    loop_connections[3].targetItem = items[0]
    loop_connections[3].targetNode = items[0].nodes[0]
    with pytest.raises(pt.PipingTraversalException):
        list(
            pt.traverse_items_and_connections(
                items, loop_connections, connections[0], lambda x: x == connections[4]
            )
        )

    # Case 7: Multiple paths to same destination
    fork_connections = copy.deepcopy(connections)
    fork_connections[3].targetItem = items[2]
    fork_connections[3].targetNode = items[2].nodes[0]
    with pytest.raises(pt.PipingTraversalException):
        list(
            pt.traverse_items_and_connections(
                items, fork_connections, connections[0], lambda x: x == connections[4]
            )
        )


def test_sort_connected_items_and_connections():
    """Test sorting connections and items of a piping network segment."""

    # Test normal case starting with a connection
    # Create test items and connections
    def make_items_and_connections():
        items = [piping.BallValve(nodes=[piping.PipingNode() for i in range(2)]) for _ in range(3)]
        connections = [piping.Pipe() for _ in range(3)]
        # Set up connections in a shuffled order
        connections[0].sourceItem = items[1]
        connections[0].targetItem = items[2]
        connections[1].sourceItem = items[0]
        connections[1].targetItem = items[1]
        connections[2].sourceItem = None  # First connection has no source
        connections[2].targetItem = items[0]
        return items, connections

    items, connections = make_items_and_connections()

    # Sort connections and items
    sorted_items, sorted_connections = pt.sort_connected_items_and_connections(items, connections)

    # Check if sorting is correct
    assert sorted_connections[0] == connections[2]  # First connection (no source)
    assert sorted_connections[1] == connections[1]  # Middle connection
    assert sorted_connections[2] == connections[0]  # Last connection

    assert sorted_items[0] == items[0]  # First item
    assert sorted_items[1] == items[1]  # Middle item
    assert sorted_items[2] == items[2]  # Last item

    # Check if ordering again gives the same result
    sorted_items2, sorted_connections2 = pt.sort_connected_items_and_connections(
        sorted_items, sorted_connections
    )

    assert sorted_items2 == sorted_items
    assert sorted_connections2 == sorted_connections

    # Test normal case starting with an item
    # Add a source item to the first connection
    items, connections = make_items_and_connections()
    new_item = piping.BallValve(nodes=[piping.PipingNode() for i in range(2)])
    connections[2].sourceItem = new_item
    items.append(new_item)

    # Sort connections and items again
    sorted_items, sorted_connections = pt.sort_connected_items_and_connections(items, connections)

    # Check if sorting is correct
    assert sorted_connections[0] == connections[2]  # First connection (with source)
    assert sorted_connections[1] == connections[1]  # Middle connection
    assert sorted_connections[2] == connections[0]  # Last connection

    assert sorted_items[0] == new_item  # New item
    assert sorted_items[1] == items[0]  # First item
    assert sorted_items[2] == items[1]  # Middle item
    assert sorted_items[3] == items[2]  # Last item

    # Test error cases
    # Case 1: Multiple starting points
    items, bad_connections = make_items_and_connections()
    bad_connections[1].sourceItem = None
    with pytest.raises(ValueError):
        pt.sort_connected_items_and_connections(items, bad_connections)

    # Case 2: Disconnected elements
    items, bad_connections = make_items_and_connections()
    bad_connections[0].targetItem = None
    with pytest.raises(ValueError):
        pt.sort_connected_items_and_connections(items, bad_connections)

    # Case 3: Loop in connections
    items, loop_connections = make_items_and_connections()
    loop_connections[0].targetItem = items[0]
    with pytest.raises(ValueError):
        pt.sort_connected_items_and_connections(items, loop_connections)


def test_sort_segment_connections_and_items(simple_pns_factory):
    """Test sorting connections and items within a piping network segment."""
    shuffled_pns = simple_pns_factory()

    # Shuffle the lists (but keep the connections valid)
    original_items = copy.copy(shuffled_pns.items)
    original_connections = copy.copy(shuffled_pns.connections)

    if len(shuffled_pns.items) > 1:
        shuffled_pns.items.reverse()
    if len(shuffled_pns.connections) > 1:
        shuffled_pns.connections.reverse()

    # Sort the segment
    pt.sort_segment_items_and_connections(shuffled_pns)

    # Check if items and connections are in the original order
    assert shuffled_pns.items == original_items
    assert shuffled_pns.connections == original_connections

    # Verify that the segment is still valid after sorting
    assert pt.piping_network_segment_validity_check(shuffled_pns)[0] == pt.PipingValidityCode.VALID


def test_piping_network_segment_validity_check(simple_pns_factory):
    """test some invalid piping network segments and their validity status. For
    this, use some invalid cases. Comparisons done via
    piping_toolkig.PipingValidityCode
    """

    def check_pns(segment: piping.PipingNetworkSegment):
        """Quick local namespace function for the
        piping_network_segment_validity_check"""
        return pt.piping_network_segment_validity_check(segment)

    # Case 1: Single pipe segment where pipe doesnt have the same source
    a_valve = piping.BallValve(nodes=[piping.PipingNode() for i in range(2)])
    invalid_segment = piping.PipingNetworkSegment(sourceItem=a_valve, connections=[piping.Pipe()])
    assert check_pns(invalid_segment)[0] == pt.PipingValidityCode.CONNECTION_CONVENTION_VIOLATION

    # Case 2: Segment where two pipes have the same source as pns
    invalid_segment = piping.PipingNetworkSegment(
        sourceItem=a_valve,
        connections=[piping.Pipe(sourceItem=a_valve) for i in range(2)],
    )
    assert check_pns(invalid_segment)[0] == pt.PipingValidityCode.INTERNAL_VIOLATION

    # Case 3: Segment where two pipes have the same target as pns
    invalid_segment = simple_pns_factory()  # new instance
    del invalid_segment.connections[1]
    assert check_pns(invalid_segment)[0] == pt.PipingValidityCode.INTERNAL_VIOLATION

    # Case 4: Add an unconnected item or connection
    invalid_segment = simple_pns_factory()
    invalid_segment.connections.append(piping.Pipe())
    assert check_pns(invalid_segment)[0] == pt.PipingValidityCode.INTERNAL_VIOLATION

    invalid_segment = simple_pns_factory()
    invalid_segment.items.append(piping.BallValve())
    assert check_pns(invalid_segment)[0] == pt.PipingValidityCode.INTERNAL_VIOLATION

    # Case 5: Order violation
    invalid_segment = simple_pns_factory()
    temp_item = invalid_segment.items[1]
    invalid_segment.items[1] = invalid_segment.items[0]
    invalid_segment.items[0] = temp_item
    assert check_pns(invalid_segment)[0] == pt.PipingValidityCode.INTERNAL_VIOLATION

    invalid_segment = simple_pns_factory()
    temp_connection = invalid_segment.connections[1]
    invalid_segment.connections[1] = invalid_segment.connections[0]
    invalid_segment.connections[0] = temp_connection
    assert check_pns(invalid_segment)[0] == pt.PipingValidityCode.INTERNAL_VIOLATION

    # Case 6: Node mismatch
    invalid_segment = simple_pns_factory()
    stray_node = piping.PipingNode()
    invalid_segment.connections[1].sourceNode = stray_node
    assert check_pns(invalid_segment)[0] == pt.PipingValidityCode.INTERNAL_VIOLATION
    invalid_segment = simple_pns_factory()
    invalid_segment.connections[1].targetItem = None
    assert check_pns(invalid_segment)[0] == pt.PipingValidityCode.INTERNAL_VIOLATION
    invalid_segment = simple_pns_factory()
    stray_node = piping.PipingNode()
    invalid_segment.targetNode = stray_node
    assert check_pns(invalid_segment)[0] == pt.PipingValidityCode.INTERNAL_VIOLATION

    # Final valid case
    assert check_pns(simple_pns_factory())[0] == pt.PipingValidityCode.VALID



"""
JSON Loader module for DEXPI model serialization and deserialization.

This module provides functionality to serialize DEXPI models to JSON format
and deserialize JSON data back into DEXPI model objects. It includes classes
for encoding models to dictionaries and decoding dictionaries to models.
"""

import json
from collections.abc import Callable
from datetime import datetime
from pathlib import Path
from typing import Any

import pydexpi.toolkits.base_model_utils as bmt
from pydexpi.dexpi_classes.pydantic_classes import DexpiBaseModel, DexpiDataTypeBaseModel
from pydexpi.loaders.serializer import Serializer
from pydexpi.toolkits.base_model_utils import (
    get_composition_attributes,
    get_data_attributes,
    get_reference_attributes,
)


class JsonSerializer(Serializer):
    """
    JSON implementation of the Serializer interface for DEXPI models.

    This class provides methods to save DEXPI models to JSON files and load
    DEXPI models from JSON files. It handles the conversion between DEXPI model
    objects and JSON-serializable dictionaries.
    """

    def __init__(self) -> None:
        """Initialize a new JSONLoader by constructing encoder and decoder objects as attributes."""
        super().__init__()
        self.encoder = DexpiToDictEncoder()
        self.decoder = DictToDexpiDecoder()

    def save(self, model: DexpiBaseModel, dir_path: str | Path, filename: str) -> None:
        """
        Save a DEXPI model to a JSON file.

        Parameters
        ----------
        model : DexpiBaseModel
            The DEXPI model to save
        dir_path : str or Path
            Directory path where the file will be saved
        filename : str
            Name of the file without extension

        Returns
        -------
        None
        """
        # Add json ending if not explicitly given
        if not filename.endswith(".json"):
            filename = f"{filename}.json"

        path = Path(dir_path) / filename
        with open(path, "w", encoding="utf-8") as file:
            json.dump(self.model_to_dict(model), file, indent=4, ensure_ascii=False)

    def load(self, dir_path: str | Path, filename: str) -> DexpiBaseModel:
        """
        Load a DEXPI model from a JSON file.

        Parameters
        ----------
        dir_path : str or Path
            Directory path where the file is located
        filename : str
            Name of the file without extension

        Returns
        -------
        DexpiBaseModel
            The loaded DEXPI model

        Raises
        ------
        FileNotFoundError
            If the specified file does not exist
        """
        # Add json ending if not explicitly given
        if not filename.endswith(".json"):
            filename = f"{filename}.json"

        path = Path(dir_path) / filename
        if not path.exists():
            raise FileNotFoundError(f"File {path} does not exist.")

        with open(path, encoding="utf-8") as file:
            data = json.load(file)
        return self.dict_to_model(data, external_refs={})

    def model_to_dict(self, model: DexpiBaseModel) -> dict:
        """
        Convert a DEXPI model to a JSON serializable dictionary.

        Parameters
        ----------
        model : DexpiBaseModel
            The DEXPI model to convert

        Returns
        -------
        dict
            A dictionary representation of the DEXPI model
        """
        return self.encoder.dexpi_element_to_dict(model)

    def dict_to_model(self, data: dict, external_refs: dict = None) -> DexpiBaseModel:
        """
        Convert a dictionary to a DEXPI model.

        Parameters
        ----------
        data : dict
            Dictionary representation of the DEXPI model
        external_refs : dict, optional
            Dictionary of external references, by default None

        Returns
        -------
        DexpiBaseModel
            The created DEXPI model
        """
        return self.decoder.dict_to_dexpi_element(data, external_refs)


class DexpiToDictEncoder:
    """
    Encoder class for converting DEXPI models to JSON-serializable dictionaries.

    This class contains static methods to convert DEXPI model objects
    and their attributes to dictionaries that can be serialized to JSON.
    """

    @staticmethod
    def dexpi_element_to_dict(element: DexpiBaseModel | DexpiDataTypeBaseModel) -> dict:
        """
        Convert a DEXPI model element to a JSON serializable dictionary.

        Parameters
        ----------
        element : DexpiBaseModel or DexpiDataTypeBaseModel
            The element to convert

        Returns
        -------
        dict
            A dictionary representation of the element
        """

        # Get all attributes of the element by type
        raw_comp_attributes = get_composition_attributes(element)
        raw_reference_attributes = get_reference_attributes(element)
        raw_data_attributes = get_data_attributes(element)

        # Initialize dictionaries to hold the unpacked attributes for later constructor kwargs use
        comp_attribute_dict = {}
        reference_attribute_dict = {}
        data_attribute_dict = {}

        # Package composition attributes. This is done recursively with dexpi_element_to_dict
        # to ensure that nested elements are also converted to dictionaries.
        for attr, attr_val in raw_comp_attributes.items():
            comp_attribute_dict[attr] = _call_on_list_or_object_or_none(
                DexpiToDictEncoder.dexpi_element_to_dict, attr_val
            )

        # Package reference attributes. For reference attributes, we only store the IDs.
        for attr, attr_val in raw_reference_attributes.items():
            reference_attribute_dict[attr] = _call_on_list_or_object_or_none(
                lambda x: x.id, attr_val
            )

        # Package data attributes with package_data_attribute.
        for attr, attr_val in raw_data_attributes.items():
            data_attribute_dict[attr] = _call_on_list_or_object_or_none(
                DexpiToDictEncoder.package_data_attribute, attr_val
            )

        # Combine all attributes into a single dictionary. Add the type of the element and the ID if
        # it exists.
        element_dict = {"uri": element.uri}
        if isinstance(element, DexpiBaseModel):
            element_dict["id"] = element.id
        if comp_attribute_dict:
            element_dict["composition"] = comp_attribute_dict
        if reference_attribute_dict:
            element_dict["reference"] = reference_attribute_dict
        if data_attribute_dict:
            element_dict["data"] = data_attribute_dict

        return element_dict

    @staticmethod
    def package_data_attribute(
        attribute_val: DexpiDataTypeBaseModel | str | int | float | datetime,
    ) -> dict | str | int | float:
        """
        Unpack a data attribute to a JSON serializable format.

        Uses the correct method to convert the attribute value based on its type. If it is a DEXPI
        DataTypeBaseModel, it will be converted to a dictionary using dexpi_element_to_dict.
        Primitive types (str, int, float) are returned as is, and datetime objects are converted to
        strings.

        Parameters
        ----------
        attribute_val : DexpiDataTypeBaseModel or str or int or float or datetime
            The attribute value to package

        Returns
        -------
        dict or str or int or float
            JSON serializable representation of the attribute

        Raises
        ------
        TypeError
            If the attribute has an unsupported data type
        """
        if isinstance(attribute_val, DexpiDataTypeBaseModel):
            return DexpiToDictEncoder.dexpi_element_to_dict(attribute_val)
        elif isinstance(attribute_val, str | int | float):
            return attribute_val
        elif isinstance(attribute_val, datetime):
            return str(attribute_val)
        else:
            raise TypeError(f"Unsupported data type: {type(attribute_val)}")


class DictToDexpiDecoder:
    """
    Decoder class for converting dictionaries to DEXPI model objects.

    This class handles the reconstruction of DEXPI model objects from
    dictionary representations, including resolving object references.
    """

    def __init__(self) -> None:
        """Initialize a new decoder with an empty object registry.

        This registry is used to keep track of objects created during the
        compositional pass, allowing for reference resolution in the
        reference pass."""
        self.object_registry = {}

    def dict_to_dexpi_element(self, data: dict, external_refs: dict = None) -> DexpiBaseModel:
        """
        Convert a dictionary to a DEXPI model element.

        This is done in two passes:
        1. Compositional pass: Create the DEXPI model objects and their attributes and resolve
           compositional and data dependencies.
        2. Reference pass: Once all objects are created, resolve reference relationships by
           retrieving referenced objects from the object registry.

        Parameters
        ----------
        data : dict
            Dictionary representation of the DEXPI model
        external_refs : dict, optional
            Dictionary of external references, by default None

        Returns
        -------
        DexpiBaseModel
            The reconstructed DEXPI model
        """
        # Reset and set the object registry
        self.object_registry = external_refs if external_refs else {}

        # Create the object with compositional and data attributes
        the_object = self._compositional_pass(data)

        # Resolve references in the object
        self._reference_pass(data)
        return the_object

    def _compositional_pass(self, data: dict) -> DexpiBaseModel:
        """Recursively convert dictionaries to nested DEXPI model elements.

        This method handles the creation of DEXPI model objects including
        their composition and data attributes but not references.

        Parameters
        ----------
        data : dict
            Dictionary representation of the DEXPI model

        Returns
        -------
        DexpiBaseModel
            The constructed DEXPI model without resolved references
        """

        # Retrieve the model class from the Dexpi classes
        model_uri = data.get("uri")
        model_class = bmt.get_dexpi_class_from_uri(model_uri)

        raw_comp_attrs = data.get("composition", {})

        # Prepare the composition attributes for the model
        comp_attr_args = {}
        for attr, attr_val in raw_comp_attrs.items():
            comp_attr_args[attr] = _call_on_list_or_object_or_none(
                self._compositional_pass, attr_val
            )

        # Prepare the data attributes for the model

        raw_data_attrs = data.get("data", {})

        data_attr_args = {}
        for attr, attr_val in raw_data_attrs.items():
            if isinstance(attr_val, list):
                data_attr_args[attr] = [self._unpack_data_attribute(item) for item in attr_val]
            else:
                data_attr_args[attr] = self._unpack_data_attribute(attr_val)

        # Construct the new object with the model class and attributes by stacking the arguments
        model_id = data.get("id")
        model_args = {"id": model_id} if model_id else {}
        model_args.update(comp_attr_args)
        model_args.update(data_attr_args)

        # Create an instance of the model class with the arguments
        new_object = model_class(**model_args)

        # Register the new object in the object registry for the reference pass if there is an ID
        # If there isn't, then the object cannot be referenced, so it doesn't need to be registered.
        if model_id is not None:
            self.object_registry[new_object.id] = new_object

        return new_object

    def _unpack_data_attribute(self, attribute_val: dict | str | int | float) -> Any:
        """
        Unpack data attributes from a dictionary to the appropriate type.

        Parameters
        ----------
        attribute_val : dict or str or int or float
            The attribute value to unpack

        Returns
        -------
        Any
            The unpacked value, either a DEXPI model or primitive type
        """

        # If data type base model, unpack as a DexpiBaseModel element
        if isinstance(attribute_val, dict):
            return self._compositional_pass(attribute_val)
        # Else, return as is if it's a primitive type
        else:
            return attribute_val

    def _reference_pass(self, data: dict) -> None:
        """
        Resolve references in the data dictionary.

        This method resolves object references using the object registry
        after all objects have been created in the compositional pass.

        Parameters
        ----------
        data : dict
            Dictionary representation of the DEXPI model

        Returns
        -------
        None
        """

        # Retrieve the model class from the Dexpi classes
        object_id = data.get("id")

        the_object = self.object_registry.get(object_id)

        raw_reference_attrs = data.get("reference", {})

        for attr, attr_val in raw_reference_attrs.items():
            resolved_refs = _call_on_list_or_object_or_none(
                self.object_registry.__getitem__, attr_val
            )
            setattr(the_object, attr, resolved_refs)

        # Call on all composition in the hierarchy
        raw_comp_attrs = data.get("composition", {})

        for attr, attr_val in raw_comp_attrs.items():
            _call_on_list_or_object_or_none(self._reference_pass, attr_val)


def _call_on_list_or_object_or_none(func: Callable, obj: Any) -> Any:
    """
    Call a function on an object, list of objects, or return None if the object is None.

    This helper function provides a unified interface for applying a function to
    either a single object or a list of objects, handling None values appropriately.

    Parameters
    ----------
    func : Callable
        The function to call on the object.
    obj : Any
        The object to process, which can be a list, a single object, or None.

    Returns
    -------
    Any
        The result of the function call on the object, a list of results if the input
        was a list, or None if the input was None.
    """
    if obj is None:
        return None
    elif isinstance(obj, list):
        return [func(item) for item in obj]
    else:
        return func(obj)



import copy
import datetime

import pytest

from pydexpi.dexpi_classes import dexpiModel, equipment, metaData, piping
from pydexpi.loaders import proteus_serializer
from pydexpi.toolkits import piping_toolkit as pt


@pytest.fixture()
def loaded_example_dexpi():
    """Initialize DEXPI loader."""
    path = "data"
    filename = "C01V04-VER.EX01"
    serializer = proteus_serializer.ProteusSerializer()
    example_dexpi = serializer.load(path, filename)
    return example_dexpi


@pytest.fixture()
def simple_pns_factory():
    """Piping network segment with two pipes and valves"""

    def _create_simple_pns(no_valves=3):
        """Create a simple piping network segment with two pipes and valves."""
        valves = [
            piping.BallValve(nodes=[piping.PipingNode() for i in range(2)])
            for j in range(no_valves)
        ]
        pipes = [piping.Pipe() for i in range(no_valves)]
        segment = pt.construct_new_segment(
            valves,
            pipes,
            target_connector_item=valves[-1],
            target_connector_node_index=1,
        )
        return segment

    return _create_simple_pns


@pytest.fixture()
def simple_conceptual_model_factory(simple_pns_factory):
    """Simple conceptual model with one tank, two piping network systems for
    inlet and outlet, and an empty metadata."""

    def _create_simple_conceptual_model():
        """Create a simple conceptual model with one tank and two piping network systems."""
        the_equipment = equipment.Tank(nozzles=[equipment.Nozzle() for i in range(2)])
        pns1 = simple_pns_factory()
        pns2 = copy.deepcopy(pns1)
        pt.append_connection_to_unconnected_segment(pns1, piping.Pipe(), -1)
        pt.append_connection_to_unconnected_segment(pns2, piping.Pipe(), -1)
        pt.connect_piping_network_segment(pns1, the_equipment.nozzles[0])
        pt.connect_piping_network_segment(pns2, the_equipment.nozzles[1], as_source=True)
        the_systems = [piping.PipingNetworkSystem(segments=[i]) for i in [pns1, pns2]]
        the_metadata = metaData.MetaData()
        the_conceptual_model = dexpiModel.ConceptualModel(
            pipingNetworkSystems=the_systems,
            taggedPlantItems=[the_equipment],
            metaData=the_metadata,
        )
        return the_conceptual_model

    return _create_simple_conceptual_model


@pytest.fixture()
def simple_dexpi_model_factory(simple_conceptual_model_factory):
    """Simple dexpi model containing a simple conceptual model fixure and some
    data attributes"""

    def _create_simple_dexpi_model():
        """Create a simple dexpi model with a simple conceptual model."""
        conceptual_model = (
            simple_conceptual_model_factory()
        )  # call factory to create a new instance
        data_dict = {
            "exportDateTime": datetime.datetime(1990, 10, 3),
            "originatingSystemName": "The System Name",
            "originatingSystemVendorName": "The System Vendor",
            "originatingSystemVersion": "The System Version",
        }
        the_dexpi_model = dexpiModel.DexpiModel(conceptualModel=conceptual_model, **data_dict)
        return the_dexpi_model

    return _create_simple_dexpi_model



"""Connector renaming convention for unique labeling across patterns."""

from collections import defaultdict

from pydexpi.syndata.pattern import Connector, Pattern


class ConnectorRenamingConvention:
    """Class to manage the renaming of connectors in a pattern to generate unique labels across
    patterns."""

    def __init__(self):
        """Initialize the ConnectorRenamingConvention class."""
        # Dictionary to keep track of the number of times a pattern has been encountered.
        self.encountered_pattern_counter = defaultdict(int)

    def rename_connectors(
        self, pattern_to_rename: Pattern, connectors_to_skip: list[Connector] = None
    ) -> None:
        """Rename connectors in the given pattern to unique labels accross patterns.

        The new names are obtained from calling _make_new_connector_name() for each connector
        internally.

        Parameters
        ----------
        pattern_to_rename : Pattern
            The pattern whose connectors are to be renamed.
        connectors_to_skip : list[Connector], optional
            A list of connectors to skip renaming.

        Returns
        -------
        None: The pattern is modified in place.
        """
        connectors_to_skip = connectors_to_skip or []
        for connector in list(pattern_to_rename.connectors.values()):
            if connector in connectors_to_skip:
                continue
            else:
                new_label = self._make_new_connector_name(connector, pattern_to_rename)
                pattern_to_rename.relabel_connector(connector, new_label)

        # Increment the counter for the encountered pattern
        self.encountered_pattern_counter[pattern_to_rename.label] += 1

    def _make_new_connector_name(self, the_connector: Connector, parent_pattern: Pattern) -> str:
        """Create a new connector name based on the parent pattern's name and the connector's type.

        The new name is generated by combining the parent pattern's label, the connector's type,
        and a counter. This ensures unique labeling accross patterns.

        Parameters
        ----------
        the_connector : Connector
            The connector for which to create a new name.
        parent_pattern : Pattern
            The parent pattern to which the connector belongs.

        Returns
        -------
        str
            The new connector name.
        """
        current_pattern_index = self.encountered_pattern_counter[parent_pattern.label]
        return f"{parent_pattern.label}_{current_pattern_index}_{the_connector.label}"

    def reset(self) -> None:
        """Reset the counter for all encountered patterns."""
        self.encountered_pattern_counter = defaultdict(int)



"""This test module contains tests for the base_model_utils module"""

import pytest

from pydexpi.dexpi_classes.pydantic_classes import ActuatingSystem, Equipment, PipingNetworkSegment
from pydexpi.toolkits import base_model_utils as bmt


def test_get_composition_attributes(simple_pns_factory):
    """Test function to retrieve composition attributes."""
    # Define the expected set of composition attributes
    attributes = bmt.get_composition_attributes(simple_pns_factory())
    # Check if all composition attributes were found
    comp_attrs = {"connections", "items", "customAttributes"}
    assert set(attributes) == comp_attrs


def test_get_reference_attributes(simple_pns_factory):
    """Test function to retrieve reference attributes."""
    attributes = bmt.get_reference_attributes(simple_pns_factory())
    # Define the expected set of reference attributes
    ref_attrs = {"sourceNode", "sourceItem", "targetNode", "targetItem"}
    # Check if all reference attributes were found
    assert set(attributes) == ref_attrs


def test_get_data_attributes(simple_pns_factory):
    """Test function to retrieve data attributes."""
    attributes = bmt.get_data_attributes(simple_pns_factory())
    # Define the expected set of data attributes
    data_attribute_names = {
        "colorCode",
        "flowDirection",
        "fluidCode",
        "heatTracingType",
        "heatTracingTypeRepresentation",
        "inclination",
        "insulationThickness",
        "insulationType",
        "jacketedPipe",
        "lowerLimitHeatTracingTemperature",
        "nominalDiameterNumericalValueRepresentation",
        "nominalDiameterRepresentation",
        "nominalDiameterStandard",
        "nominalDiameterTypeRepresentation",
        "onHold",
        "operatingTemperature",
        "pipingClassCode",
        "pressureTestCircuitNumber",
        "primarySecondaryPipingNetworkSegment",
        "segmentNumber",
        "siphon",
        "slope",
    }
    # Check if all data attributes were found
    assert set(attributes) == data_attribute_names


def test_get_attributes_with_category(simple_pns_factory):
    """Test the function get get_attributes_with_category, though mostly already
    covered by implementation functions with data, composition, and reference.
    """
    assert bmt._get_attributes_with_category(simple_pns_factory(), "data")
    assert not bmt._get_attributes_with_category(simple_pns_factory(), "something else")


def test_get_dexpi_class() -> None:
    """Test the function get_dexpi_class."""
    assert bmt.get_dexpi_class("PipingNetworkSegment") == PipingNetworkSegment
    assert bmt.get_dexpi_class("Equipment") == Equipment
    assert bmt.get_dexpi_class("ActuatingSystem") == ActuatingSystem

    with pytest.raises(AttributeError):
        bmt.get_dexpi_class("NonExistentClass")


def test_get_dexpi_class_from_uri() -> None:
    pns = PipingNetworkSegment()
    assert bmt.get_dexpi_class_from_uri(pns.uri) == PipingNetworkSegment

    equipment = Equipment()
    assert bmt.get_dexpi_class_from_uri(equipment.uri) == Equipment

    actuating_system = ActuatingSystem()
    assert bmt.get_dexpi_class_from_uri(actuating_system.uri) == ActuatingSystem

    with pytest.raises(AttributeError):
        bmt.get_dexpi_class_from_uri("http://example.com/NonExistentClass")



from pydexpi.loaders.proteus_serializer import ProteusSerializer
from pydexpi.loaders.proteus_serializer.core import ErrorLevels
from pydexpi.toolkits import base_model_utils as bmu, model_toolkit as mt, piping_toolkit as pt


def test_parse_proteus_to_dexpi():
    """Parse proteus root tree to dexpi classes."""
    path = "data"
    filename = "C01V04-VER.EX01"
    serializer = ProteusSerializer()
    example_dexpi = serializer.load(path, filename)
    # Check if total number of objects is still correct (for equipment, piping,
    # and instrumentation)
    assert len(mt.get_all_instances_in_model(example_dexpi)) == 214
    # Check the piping network segments
    for system in example_dexpi.conceptualModel.pipingNetworkSystems:
        for segment in system.segments:
            assert (
                pt.piping_network_segment_validity_check(segment)[0] == pt.PipingValidityCode.VALID
            )
    # Run some misc tests
    assert len(example_dexpi.conceptualModel.actuatingSystems) == 3
    assert (
        example_dexpi.conceptualModel.taggedPlantItems[0].__class__.__name__ == "PlateHeatExchanger"
    )

    # Assert that there are only ErrorLevels.INFO in the parser
    error_registry = serializer.proteus_loader.error_registry
    non_info_errors = error_registry.get_errors(
        [ErrorLevels.ERROR, ErrorLevels.WARNING, ErrorLevels.CRITICAL]
    )
    assert len(non_info_errors) == 0


def test_parse_generic_attributes(loaded_example_dexpi):
    """Test if DEXPI generic attributes are parsed correctly from proteus."""
    dexpi_model = loaded_example_dexpi
    assert dexpi_model.conceptualModel.taggedPlantItems[0].plateHeight.unit.value == "mm"
    assert float(dexpi_model.conceptualModel.taggedPlantItems[0].plateHeight.value) == 850
    assert (
        dexpi_model.conceptualModel.taggedPlantItems[4]
        .nozzles[0]
        .nodes[0]
        .nominalDiameterRepresentation
        == "DN 80"
    )


def test_parse_all_attributes(loaded_example_dexpi):
    """Count all None attributes in the loaded dexpi model and make sure they are the expected amount."""
    all_instances = mt.get_all_instances_in_model(loaded_example_dexpi)

    # Count all attributes that are None
    attr_count = 0
    not_none_fields = 0
    for instance in all_instances:
        raw_attrs = bmu.get_data_attributes(instance)
        for fld_value in raw_attrs.values():
            if fld_value is not None:
                not_none_fields += 1
                if isinstance(fld_value, list):
                    attr_count += len(fld_value)
                else:
                    attr_count += 1

    assert attr_count == 732
    assert not_none_fields == 732



from decimal import Decimal

import pytest

from pydexpi.dexpi_classes import (
    dexpiModel,
    equipment,
    physicalQuantities,
)
from pydexpi.loaders.serializer import PickleSerializer


@pytest.fixture
def reactor_model():
    """Creates a simple reactor DEXPI model.

    Returns
    -------
    nx.DiGraph
        NetworkX graph of a reactor dexpi model.
    """
    return dexpiModel.DexpiModel(
        conceptualModel=dexpiModel.ConceptualModel(
            taggedPlantItems=[
                equipment.PressureVessel(
                    tagName="R-1",
                    chambers=[
                        equipment.Chamber(
                            upperLimitDesignPressure=physicalQuantities.PressureGauge(
                                unit=physicalQuantities.PressureGaugeUnit("bar"),
                                value=Decimal(1),
                            )
                        )
                    ],
                )
            ]
        )
    )


def test_pickle_serializer(reactor_model, tmp_path):
    """Saves and loads a dummy reactor model and compares it with the original
    object.

    Parameters
    ----------
    reactor_model : DexpiModel
        Dummy reactor model.
    """
    my_serializer = PickleSerializer()
    my_serializer.save(reactor_model, tmp_path, "test_pkl_serializer")
    new_reactor_model = my_serializer.load(tmp_path, "test_pkl_serializer")
    assert reactor_model == new_reactor_model



"""Tests for the parsing_utils module of the proteus serializer."""

from unittest.mock import Mock

import pytest

from pydexpi.dexpi_classes.pydantic_classes import DexpiBaseModel
from pydexpi.loaders.proteus_serializer.core import ErrorLevels, ParserModule
from pydexpi.loaders.proteus_serializer.parsing_utils import (
    ErrorTemplates,
    add_by_inferring_type,
    filter_none,
    is_associated_with,
    redirect_errors_to_registry,
)


class TestErrorTemplates:
    """Test the ErrorTemplates class and its static methods."""

    def test_id_not_found(self) -> None:
        """Test the id_not_found error template."""
        element_name = "TestElement"
        message, level = ErrorTemplates.id_not_found(element_name)

        expected_message = f"ID not found for element '{element_name}'. Element Skipped."
        assert message == expected_message
        assert level == ErrorLevels.ERROR

    def test_skip_pass(self) -> None:
        """Test the skip_pass error template."""
        pass_name = "reference"
        element_name = "TestElement"
        message, level = ErrorTemplates.skip_pass(pass_name, element_name)

        expected_message = (
            f"'{pass_name}' pass skipped for element '{element_name}' due to error in previous "
            f"pass. Element Skipped."
        )
        assert message == expected_message
        assert level == ErrorLevels.INFO

    def test_assoc_adding_error(self) -> None:
        """Test the assoc_adding_error error template."""
        element_name = "TestElement"
        associated_obj_id = "AssocObj123"
        message, level = ErrorTemplates.assoc_adding_error(element_name, associated_obj_id)

        expected_message = (
            f"Object {associated_obj_id} could not be added as reference to {element_name}."
        )
        assert message == expected_message
        assert level == ErrorLevels.ERROR

    def test_no_assoc_ctrl(self) -> None:
        """Test the no_assoc_ctrl error template."""
        element_name = "TestElement"
        associated_obj_id = "AssocObj123"
        message, level = ErrorTemplates.no_assoc_ctrl(element_name, associated_obj_id)

        expected_message = (
            f"Association of {element_name} with {associated_obj_id} is missing in control pass. "
            f"Attempt adding association."
        )
        assert message == expected_message
        assert level == ErrorLevels.WARNING

    def test_assoc_added_ctrl_normal(self) -> None:
        """Test the assoc_added_ctrl error template with normal order."""
        element_name = "TestElement"
        associated_obj_id = "AssocObj123"
        field_name = "test_field"
        message, level = ErrorTemplates.assoc_added_ctrl(
            element_name, associated_obj_id, field_name
        )

        expected_message = (
            f"Association of {element_name} with {associated_obj_id} added in control pass to "
            f"field {field_name}."
        )
        assert message == expected_message
        assert level == ErrorLevels.INFO

    def test_assoc_added_ctrl_reverse(self) -> None:
        """Test the assoc_added_ctrl error template with reverse order."""
        element_name = "TestElement"
        associated_obj_id = "AssocObj123"
        field_name = "test_field"
        message, level = ErrorTemplates.assoc_added_ctrl(
            element_name, associated_obj_id, field_name, reverse=True
        )

        expected_message = (
            f"Association of {element_name} with {associated_obj_id} added in control pass in "
            f"reverse order to field {field_name}."
        )
        assert message == expected_message
        assert level == ErrorLevels.INFO

    def test_assoc_not_added_ctrl(self) -> None:
        """Test the assoc_not_added_ctrl error template."""
        element_name = "TestElement"
        associated_obj_id = "AssocObj123"
        message, level = ErrorTemplates.assoc_not_added_ctrl(element_name, associated_obj_id)

        expected_message = (
            f"Association of {element_name} with {associated_obj_id} could not be added in control "
            f"pass."
        )
        assert message == expected_message
        assert level == ErrorLevels.ERROR

    def test_inval_assoc_type(self) -> None:
        """Test the inval_assoc_type error template."""
        element_name = "TestElement"
        assoc_type = "InvalidType"
        message, level = ErrorTemplates.inval_assoc_type(element_name, assoc_type)

        expected_message = (
            f"Invalid association type '{assoc_type}' for {element_name}. Association skipped."
        )
        assert message == expected_message
        assert level == ErrorLevels.ERROR


class MockAssociatedModel(DexpiBaseModel):
    """Another mock DEXPI model for association testing."""

    description: str | None = None


class MockAssociatedModel2(DexpiBaseModel):
    """Another mock DEXPI model for association testing."""

    description: str | None = None


class MockDexpiModel(DexpiBaseModel):
    """Mock DEXPI model for testing purposes."""

    name: str | None = None
    value: int | None = None
    items: list[str] | None = None
    other_model: MockAssociatedModel | None = None
    model_list: list[MockAssociatedModel2] | None = None


class MockIncompatibleModel(DexpiBaseModel):
    """A mock model that's incompatible with MockDexpiModel fields."""

    incompatible_field: str = "incompatible"


class TestAddByInferringType:
    """Test the add_by_inferring_type function."""

    def test_add_to_none_field_success(self) -> None:
        """Test successfully adding to a None field."""
        container = MockDexpiModel()
        to_add = MockAssociatedModel(description="test")

        # Initially other_model should be None
        assert container.other_model is None

        field_name = add_by_inferring_type(to_add, container)

        # The function should find a suitable field and return its name
        assert field_name == "other_model"
        # Since we can't predict which field it chooses, just verify one was set
        assert container.other_model is not None or container.model_list is not None

    def test_add_to_list_field_success(self) -> None:
        """Test successfully adding to a list field."""
        container = MockDexpiModel(model_list=[])
        to_add = MockAssociatedModel2(description="test")

        field_name = add_by_inferring_type(to_add, container)

        # The function should find a suitable field and add the item
        assert field_name == "model_list"
        # Verify the item was added to one of the compatible fields
        assert to_add in container.model_list

    def test_add_to_existing_list_field(self) -> None:
        """Test adding to an existing list with items."""
        existing_item = MockAssociatedModel2(description="existing")
        container = MockDexpiModel(model_list=[existing_item])
        to_add = MockAssociatedModel2(description="new")

        field_name = add_by_inferring_type(to_add, container)

        # The function should find a suitable field
        assert field_name is not None
        # The existing item should still be there
        assert existing_item in container.model_list
        # The new item should be added somewhere
        assert to_add in container.model_list

    def test_no_suitable_field_raises_error(self) -> None:
        """Test that ValueError is raised when no suitable field is found."""
        # Create a model with all fields filled
        first_model = MockAssociatedModel(description="existing")
        container = MockDexpiModel(
            name="test",
            value=42,
            items=["item1"],
            other_model=first_model,
            model_list=[MockAssociatedModel2(description="existing_list_item")],
        )

        # Try to add something that doesn't fit anywhere - incompatible type
        to_add = MockIncompatibleModel(incompatible_field="test")

        with pytest.raises(ValueError, match="No suitable field found"):
            add_by_inferring_type(to_add, container)

        # Assert initial state is unchanged
        assert container.other_model == first_model

    def test_validation_error_handling(self) -> None:
        """Test that ValidationError is handled gracefully."""
        container = MockDexpiModel()

        # Create an object that should work with the container
        to_add = MockAssociatedModel(description="test")

        # This should work despite potential validation errors
        field_name = add_by_inferring_type(to_add, container)
        assert field_name is not None


class TestIsAssociatedWith:
    """Test the is_associated_with function."""

    def test_direct_association(self) -> None:
        """Test detection of direct association."""
        member = MockAssociatedModel(description="member")
        container = MockDexpiModel(other_model=member)

        assert is_associated_with(member, container) is True

    def test_list_association(self) -> None:
        """Test detection of association in a list."""
        member = MockAssociatedModel2(description="member")
        other_member = MockAssociatedModel2(description="other")
        container = MockDexpiModel(model_list=[other_member, member])

        assert is_associated_with(member, container) is True

    def test_no_association(self) -> None:
        """Test when there is no association."""
        member = MockAssociatedModel(description="member")
        other_member = MockAssociatedModel(description="other")
        container = MockDexpiModel(other_model=other_member)

        assert is_associated_with(member, container) is False

    def test_empty_container(self) -> None:
        """Test with empty container."""
        member = MockAssociatedModel(description="member")
        container = MockDexpiModel()

        assert is_associated_with(member, container) is False

    def test_empty_list_field(self) -> None:
        """Test with empty list field."""
        member = MockAssociatedModel(description="member")
        container = MockDexpiModel(model_list=[])

        assert is_associated_with(member, container) is False

    def test_none_fields_ignored(self) -> None:
        """Test that None fields are handled properly."""
        member = MockAssociatedModel(description="member")
        container = MockDexpiModel(other_model=None, model_list=None)

        assert is_associated_with(member, container) is False


class TestFilterNone:
    """Test the filter_none function."""

    def test_filter_empty_list(self) -> None:
        """Test filtering an empty list."""
        result = filter_none([])
        assert result == []

    def test_filter_no_none_values(self) -> None:
        """Test filtering a list with no None values."""
        items = [1, 2, "test", True]
        result = filter_none(items)
        assert result == items

    def test_filter_all_none_values(self) -> None:
        """Test filtering a list with all None values."""
        items = [None, None, None]
        result = filter_none(items)
        assert result == []

    def test_filter_mixed_values(self) -> None:
        """Test filtering a list with mixed values."""
        items = [1, None, "test", None, 42, None, True]
        expected = [1, "test", 42, True]
        result = filter_none(items)
        assert result == expected

    def test_filter_preserves_order(self) -> None:
        """Test that filtering preserves the original order."""
        items = ["first", None, "second", None, "third"]
        expected = ["first", "second", "third"]
        result = filter_none(items)
        assert result == expected

    def test_filter_with_falsy_values(self) -> None:
        """Test that falsy values (but not None) are preserved."""
        items = [0, None, "", None, False, None, []]
        expected = [0, "", False, []]
        result = filter_none(items)
        assert result == expected


class MockLoaderModule(ParserModule):
    """Mock LoaderModule for testing the decorator."""

    def __init__(self) -> None:
        # Mock the context and register_error method
        self.context = Mock()
        self.register_error = Mock()

    def compositional_pass(self) -> None:
        """Mock implementation."""
        pass


class TestRedirectErrorsToRegistry:
    """Test the redirect_errors_to_registry decorator."""

    def test_successful_execution(self) -> None:
        """Test that successful function execution works normally."""

        @redirect_errors_to_registry
        def test_function(self, x: int, y: int) -> int:
            return x + y

        loader = MockLoaderModule()
        result = test_function(loader, 5, 3)

        assert result == 8
        loader.register_error.assert_not_called()

    def test_exception_redirected_to_registry(self) -> None:
        """Test that exceptions are redirected to the error registry."""
        test_exception = ValueError("Test error")

        @redirect_errors_to_registry
        def test_function(self, x: int) -> int:
            raise test_exception

        loader = MockLoaderModule()
        result = test_function(loader, 5)

        assert result is None
        loader.register_error.assert_called_once()

        # Check the arguments passed to register_error
        call_args = loader.register_error.call_args
        assert "Test error" in call_args.kwargs["message"]
        assert call_args.kwargs["level"] == ErrorLevels.ERROR
        assert call_args.kwargs["exception"] == test_exception

    def test_no_args_raises_error(self) -> None:
        """Test that decorator raises error when called with no arguments."""

        @redirect_errors_to_registry
        def test_function() -> None:
            raise ValueError("Should not reach here")

        with pytest.raises(ValueError, match="first argument 'self' must be an instance"):
            test_function()

    def test_wrong_self_type_raises_error(self) -> None:
        """Test that decorator raises error when self is not LoaderModule."""

        @redirect_errors_to_registry
        def test_function(self) -> None:
            raise ValueError("Should not reach here")

        wrong_self = "not_a_loader_module"

        with pytest.raises(ValueError, match="first argument 'self' must be an instance"):
            test_function(wrong_self)

    def test_message_formatting(self) -> None:
        """Test that error message is formatted correctly."""

        @redirect_errors_to_registry
        def test_parsing_step(self) -> None:
            raise KeyError("Missing key")

        loader = MockLoaderModule()
        loader.__class__.__name__ = "TestLoader"

        test_parsing_step(loader)

        call_args = loader.register_error.call_args
        message = call_args.kwargs["message"]

        assert "test_parsing_step" in message
        assert "TestLoader" in message
        assert "Missing key" in message

    def test_function_name_in_error_message(self) -> None:
        """Test that the function name appears in the error message."""

        @redirect_errors_to_registry
        def specific_parsing_function(self) -> None:
            raise IndexError("Index out of range")

        loader = MockLoaderModule()
        specific_parsing_function(loader)

        call_args = loader.register_error.call_args
        message = call_args.kwargs["message"]

        assert "specific_parsing_function" in message

    def test_kwargs_passed_through(self) -> None:
        """Test that keyword arguments are passed through correctly."""

        @redirect_errors_to_registry
        def test_function(self, x: int, y: int = 10, z: str = "default") -> str:
            return f"{x}-{y}-{z}"

        loader = MockLoaderModule()
        result = test_function(loader, 5, y=20, z="custom")

        assert result == "5-20-custom"

    def test_preserves_function_metadata(self) -> None:
        """Test that the decorator preserves function metadata."""

        @redirect_errors_to_registry
        def documented_function(self) -> int:
            """This function has documentation."""
            return 42

        # The @wraps decorator should preserve the original function's metadata
        assert documented_function.__name__ == "documented_function"
        assert "This function has documentation" in documented_function.__doc__



from .pydantic_classes import (
    CustomAreaAttribute,
    CustomAttribute,
    CustomAttributeOwner,
    CustomElectricalFrequencyAttribute,
    CustomForceAttribute,
    CustomHeatTransferCoefficientAttribute,
    CustomIntegerAttribute,
    CustomLengthAttribute,
    CustomMassAttribute,
    CustomMassFlowRateAttribute,
    CustomMultiLanguageStringAttribute,
    CustomNumberPerTimeIntervalAttribute,
    CustomObject,
    CustomPercentageAttribute,
    CustomPowerAttribute,
    CustomPressureAbsoluteAttribute,
    CustomPressureGaugeAttribute,
    CustomRotationalFrequencyAttribute,
    CustomStringAttribute,
    CustomTemperatureAttribute,
    CustomVoltageAttribute,
    CustomVolumeAttribute,
    CustomVolumeFlowRateAttribute,
)



"""Conftest for synthetic data generation tests.

Contains some mock classes and package-wide fixtures"""

import pytest

from pydexpi.syndata.pattern import (
    Connector,
    Pattern,
)
from pydexpi.syndata.pattern_distribution import PatternDistribution


class DummyConnector(Connector):
    """
    A dummy implementation of the abstract `Connector` class for testing purposes.

    This class provides a minimal concrete implementation of the `_implement_connection`
    method, allowing the `Connector` class to be instantiated and tested.
    """

    def _implement_connection(self, counterpart: Connector) -> None:
        return super()._implement_connection(counterpart)

    def assess_valid_counterpart(self, counterpart: Connector) -> bool:
        if not isinstance(counterpart, DummyConnector):
            raise TypeError(f"Counterpart must be of type DummyConnector, not {type(counterpart)}")
        return True


class DummyPattern(Pattern):
    """
    A dummy implementation of the abstract `Pattern` class for testing purposes.

    This class provides a minimal concrete implementation of the `_implement_incorporation`
    method, allowing the `Pattern` class to be instantiated and tested.
    """

    def _implement_incorporation(
        self,
        counterpart: Pattern,
    ) -> None:
        return super()._implement_incorporation(counterpart)


@pytest.fixture
def simple_pattern_factory():
    """Factory function to create a simple pattern with two connectors."""

    def _factory(
        pattern_label: str,
        connector_label_prefix: str = "Connector label",
        no_pattern_connectors: int = 2,
    ) -> DummyPattern:
        """Factory function to create a simple pattern."""
        # Create primary connectors
        connectors = [
            DummyConnector(f"{connector_label_prefix}{i + 1}") for i in range(no_pattern_connectors)
        ]
        the_pattern = DummyPattern(
            pattern_label,
            connectors=connectors,
            kwinfos={"dummy_kw": "Dummy kw value"},
        )
        # Create observer pattern and add it
        observer_connectors = [
            DummyConnector(f"{connector_label_prefix}{i + 1}") for i in range(no_pattern_connectors)
        ]
        observer_pattern = DummyPattern(pattern_label, connectors=observer_connectors)
        the_pattern.add_observer("Observer tag", observer_pattern)
        return the_pattern

    return _factory


@pytest.fixture
def simple_distribution_factory(simple_pattern_factory):
    """Factory function to create a simple pattern distribution."""

    def _factory(
        name: str, no_patterns: int = 2, no_pattern_connectors: int = 2
    ) -> PatternDistribution:
        """Factory function to create a simple pattern distribution."""
        patterns = [
            simple_pattern_factory(f"pattern{i + 1}", no_pattern_connectors=no_pattern_connectors)
            for i in range(no_patterns)
        ]
        pattern_dict = {p.label: p for p in patterns}
        some_weights = [float(i) for i in range(len(pattern_dict))]
        normalized_probs = [i / sum(some_weights) for i in some_weights]
        probabilities = {p: normalized_probs[i] for i, p in enumerate(pattern_dict.keys())}
        connector_labels = patterns[0].connectors.keys()
        return PatternDistribution(name, pattern_dict, probabilities, connector_labels)

    return _factory



"""Step classes for the generator.

These classes define the steps that can be taken during the generation process."""

from abc import ABC, abstractmethod
from typing import Protocol, TypedDict

from pydexpi.syndata.connector_renaming import ConnectorRenamingConvention
from pydexpi.syndata.pattern import Connector, Pattern


class GeneratorStepDict(TypedDict):
    """Typed dictionary for the GeneratorStep class."""

    generator_step_type: str
    own_connector: str | None
    next_pattern: str | None
    next_connector: str | None
    sampled_distribution_name: str | None


class WriteableStep(Protocol):
    """Protocol for a step that can be written to a dictionary."""

    def to_dict(self) -> GeneratorStepDict:
        """Converts the step to a dictionary.

        Returns
        -------
        GeneratorStepDict
            Dictionary representation of the step.
        """


class GeneratorStep(ABC):
    """Abstract base class for a generator step."""

    @abstractmethod
    def to_dict(self) -> GeneratorStepDict:
        """Converts the GeneratorStep to a dictionary.

        Returns
        -------
        GeneratorStepDict
            Dictionary representation of the generator step.
        """

    @abstractmethod
    def execute_on(self, current_pattern: Pattern) -> None:
        """Executes the generator step.

        This method should be overridden in subclasses to provide specific functionality for each
        generator step.

        Parameters
        ----------
        current_pattern : Pattern
            The current pattern to which the generator step is applied.
        """

    @abstractmethod
    def get_termination_status(self) -> bool:
        """Checks if the generator step has reached a termination status.

        Returns
        -------
        bool
            True if the generator step has encodes a termination, False otherwise.
        """

    @abstractmethod
    def apply_renaming_convention(self, renaming_convention: ConnectorRenamingConvention) -> None:
        """Applies the renaming convention to the generator step.

        Parameters
        ----------
        renaming_convention : ConnectorRenamingConvention
            The renaming convention to apply.
        """
        pass


class AddPatternStep(GeneratorStep):
    """Generator step for selecting the next pattern.

    Attributes
    ----------
    own_connector : Connector
        The connector of the current pattern to participate in connection.
    next_pattern : Pattern
        The next pattern to be incorporated.
    next_connector : Connector
        The connector of the next pattern to be used in connection.
    sampled_distribution_name : str
        The name of the distribution that was selected for this step and that was sampled from.
    """

    def __init__(
        self,
        own_connector: Connector,
        next_pattern: Pattern,
        next_connector: Connector,
        sampled_distribution_name: str,
    ) -> None:
        """
        Initializes a generator step with the specified connectors, pattern, and distribution name.

        Parameters
        ----------
        own_connector : Connector
            The connector associated with the current step.
        next_pattern : Pattern
            The pattern to be used for the next step.
        next_connector : Connector
            The connector associated with the next step.
        sampled_distribution_name : str
            The name of the sampled distribution for this step.
        """
        super().__init__()
        if next_connector not in next_pattern.connectors.values():
            msg = (
                f"The connector {next_connector.label} is not associated "
                f"with the next pattern {next_pattern.label}."
            )
            raise ValueError(msg)
        self.own_connector = own_connector
        self.next_pattern = next_pattern
        self.next_connector = next_connector
        self.sampled_distribution_name = sampled_distribution_name

    def to_dict(self) -> GeneratorStepDict:
        """Converts the GeneratorStep to a dictionary.

        Returns
        -------
        GeneratorStepDict
            Dictionary representation of the generator step.
        """
        return {
            "generator_step_type": "add_pattern",
            "own_connector": self.own_connector.label,
            "next_pattern": self.next_pattern.label,
            "next_connector": self.next_connector.label,
            "sampled_distribution_name": self.sampled_distribution_name,
        }

    def execute_on(self, current_pattern: Pattern) -> None:
        """Executes the generator step by incorporating the next pattern.

        Parameters
        ----------
        current_pattern : Pattern
            The current pattern to which the generator step is applied.
        """
        if self.own_connector not in current_pattern.connectors.values():
            msg = (
                f"The connector {self.own_connector.label} is not associated "
                f"with the current pattern {current_pattern.label}."
            )
            raise RuntimeError(msg)
        current_pattern.incorporate_pattern(
            self.own_connector, self.next_pattern, self.next_connector
        )

    def get_termination_status(self) -> bool:
        """Checks if the generator step has reached a termination status.

        Returns
        -------
        bool
            True if the generator step has encodes a termination, False otherwise.
            Always returns False for AddPatternStep.
        """
        return False

    def apply_renaming_convention(self, renaming_convention: ConnectorRenamingConvention) -> None:
        """Applies the renaming convention to the generator step.

        Renames all connectors in the next pattern, except for the next connector (Which will be
        connected to the current pattern, thus not being part of the aggregated pattern).

        Parameters
        ----------
        renaming_convention : ConnectorRenamingConvention
            The renaming convention to apply.
        """
        renaming_convention.rename_connectors(self.next_pattern, [self.next_connector])


class InternalConnectionStep(GeneratorStep):
    """Generator step for internal connections.

    Attribute names are chosen to be consistent with the AddPatternStep class.

    Attributes
    ----------
    own_connector : Connector
        The first connector in the internal connection.
    next_connector : Connector
        The second connector in the internal connection.
    """

    def __init__(self, own_connector: Connector, next_connector: Connector):
        """
        Initializes a new instance of the class.
        Parameters
        ----------
        own_connector : Connector
            The connector associated with the current instance.
        next_connector : Connector
            The second, internal connector to be connected.

        Raises
        ------
        ValueError
            If the own_connector and next_connector are the same.
        """
        super().__init__()
        if own_connector == next_connector:
            msg = f"The connector {own_connector.label} cannot be connected to itself."
            raise ValueError(msg)
        self.own_connector = own_connector
        self.next_connector = next_connector

    def to_dict(self) -> GeneratorStepDict:
        """Converts the GeneratorStep to a dictionary.

        Returns
        -------
        GeneratorStepDict
            Dictionary representation of the generator step.
        """
        return {
            "generator_step_type": "internal_connection",
            "own_connector": self.own_connector.label,
            "next_pattern": None,
            "next_connector": self.next_connector.label,
            "sampled_distribution_name": None,
        }

    def execute_on(self, current_pattern: Pattern) -> None:
        """Executes the generator step by performing an internal connection.

        Parameters
        ----------
        current_pattern : Pattern
            The current pattern to which the generator step is applied.

        Raises
        ------
        RuntimeError
            If the connectors are not associated with the current pattern.
        """
        if self.own_connector not in current_pattern.connectors.values():
            msg = (
                f"The connector {self.own_connector.label} is not associated "
                f"with the current pattern {current_pattern.label}."
            )
            raise RuntimeError(msg)
        if self.next_connector not in current_pattern.connectors.values():
            msg = (
                f"The connector {self.next_connector.label} is not associated "
                f"with the current pattern {current_pattern.label}."
            )
            raise RuntimeError(msg)
        current_pattern.connect_internal(self.own_connector, self.next_connector)

    def get_termination_status(self) -> bool:
        """Checks if the generator step has reached a termination status.

        Returns
        -------
        bool
            True if the generator step has encodes a termination, False otherwise.
            Always returns False for InternalConnectionStep.
        """
        return False

    def apply_renaming_convention(self, renaming_convention: ConnectorRenamingConvention) -> None:
        """Applies the renaming convention to the generator step. Does nothing for this step.

        Parameters
        ----------
        renaming_convention : ConnectorRenamingConvention
            The renaming convention to apply.
        """


class TerminationStep(GeneratorStep):
    """Generator step for termination."""

    def __init__(self) -> None:
        """Initializes the TerminationStep."""
        super().__init__()

    def to_dict(self) -> GeneratorStepDict:
        """Converts the GeneratorStep to a dictionary.

        Returns
        -------
        GeneratorStepDict
            Dictionary representation of the generator step.
        """
        return {
            "generator_step_type": "termination",
            "own_connector": None,
            "next_pattern": None,
            "next_connector": None,
            "sampled_distribution_name": None,
        }

    def execute_on(self, current_pattern: Pattern) -> None:
        """Executes the termination step on the given pattern, does nothing for this step.

        Parameters
        ----------
        current_pattern : Pattern
            The current pattern to which the termination is applied.
        """

    def get_termination_status(self) -> bool:
        """Checks if the generator step has reached a termination status.

        Returns
        -------
        bool
            True if the generator step has encodes a termination, False otherwise.
            Always returns True for TerminationStep.
        """
        return True

    def apply_renaming_convention(self, renaming_convention: ConnectorRenamingConvention) -> None:
        """Applies the renaming convention to the generator step. Does nothing for this step.

        Parameters
        ----------
        renaming_convention : ConnectorRenamingConvention
            The renaming convention to apply.
        """


class InitializationStep:
    """Generator step for initialization.

    Attributes
    ----------
    next_pattern : Pattern
        The initial pattern to start the generation process.
    sampled_distribution_name : str
        The name of the distribution that was selected for this step.
    """

    def __init__(self, init_pattern: Pattern, sampled_distribution_name: str) -> None:
        """Initializes the InitializationStep.

        Parameters
        ----------
        init_pattern : Pattern
            The initial pattern to start the generation process.
        sampled_distribution_name : str
            The name of the distribution that was selected for this step.
        """
        self.init_pattern = init_pattern
        self.sampled_distribution_name = sampled_distribution_name

    def to_dict(self) -> GeneratorStepDict:
        """Converts the GeneratorStep to a dictionary.

        Returns
        -------
        GeneratorStepDict
            Dictionary representation of the generator step.
        """
        return {
            "generator_step_type": "initialization",
            "own_connector": None,
            "next_pattern": self.init_pattern.label,
            "next_connector": None,
            "sampled_distribution_name": self.sampled_distribution_name,
        }

    def get_pattern(self) -> Pattern:
        """Executes the generator step by initializing the pattern.

        Returns
        -------
        Pattern
            The initialized pattern.
        """
        return self.init_pattern


class CappingStep:
    """Step for capping a connector on a pattern.

    It incorporates a capping pattern if next_pattern and next_connectors are specified in the step.
    If none is specified, drop_connector is called.

    Attributes
    ----------
    own_connector : Connector
        The connector of the current pattern to be capped.
    next_pattern : Pattern
        The next pattern to be incorporated.
    next_connector : Connector
        The connector of the next pattern to be used in connection.
    sampled_distribution_name : str
        The name of the distribution that was selected for this step and that was sampled from.
    """

    def __init__(
        self,
        own_connector: Connector,
        next_pattern: Pattern | None = None,
        next_connector: Connector | None = None,
        sampled_distribution_name: str | None = None,
    ) -> None:
        """
        Initializes a capping step with the specified connectors, pattern, and distribution name.

        Parameters
        ----------
        own_connector : Connector
            The connector associated with the current step.
        next_pattern : Pattern, optional
            The pattern to be used for the next step.
        next_connector : Connector, optional
            The connector associated with the next step.
        sampled_distribution_name : str, optional
            The name of the sampled distribution for this step.

        Raises
        ------
        ValueError
            If the next connector is not associated with the next pattern or if the next pattern has more than one connector.
        """
        super().__init__()
        if next_pattern is not None:
            if next_connector not in next_pattern.connectors.values():
                msg = (
                    f"The connector {next_connector.label} is not associated "
                    f"with the next pattern {next_pattern.label}."
                )
                raise ValueError(msg)
            if not len(next_pattern.connectors) == 1:
                msg = f"The next pattern {next_pattern.label} has more than one connector. Invalid capping connector."
                raise ValueError(msg)
        self.own_connector = own_connector
        self.next_pattern = next_pattern
        self.next_connector = next_connector
        self.sampled_distribution_name = sampled_distribution_name

    def to_dict(self) -> GeneratorStepDict:
        """Converts the CappingStep to a dictionary.

        Returns
        -------
        GeneratorStepDict
            Dictionary representation of the generator step.
        """
        return {
            "generator_step_type": "capping",
            "own_connector": self.own_connector.label,
            "next_pattern": self.next_pattern.label if self.next_pattern else None,
            "next_connector": self.next_connector.label if self.next_connector else None,
            "sampled_distribution_name": self.sampled_distribution_name,
        }

    def execute_on(self, current_pattern: Pattern) -> None:
        """Executes the Capping step by incorporating the next pattern.

        Parameters
        ----------
        current_pattern : Pattern
            The current pattern to which the generator step is applied.
        """
        if self.next_pattern is not None:
            if self.own_connector not in current_pattern.connectors.values():
                msg = (
                    f"The connector {self.own_connector.label} is not associated "
                    f"with the current pattern {current_pattern.label}."
                )
                raise RuntimeError(msg)
            current_pattern.incorporate_pattern(
                self.own_connector, self.next_pattern, self.next_connector
            )
        else:
            current_pattern.drop_connector(self.own_connector)
